/*
 * Electric(tm) VLSI Design System
 *
 * File: edialogs.h
 * Header file for dialog handling
 * Written by: Steven M. Rubin, Electric Editor Incorporated
 *
 * Copyright (c) 1998 Electric Editor Incorporated.
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 *
 * Electric Editor Incorporated
 * 23470 Sunset Drive, Suite 108
 * Los Gatos, California 95033
 * support@electriceditor.com
 */

/*
 *    DIALOG CONTROL
 * DiaInitDialog(dia)                   initializes dialog from table "dia"
 * item = DiaNextHit()                  obtains next "item" hit in dialog processing
 * chr = DiaGetNextCharacter(&item)     obtains next character (-1 if none, -2 if item)
 * DiaDoneDialog()                      removes and terminates dialog
 *
 *    CONTROL OF STANDARD DIALOG ITEMS
 * DiaSetText(item, msg)                puts "msg" in "item" in the dialog
 * line = DiaGetText(item)              gets text in "item" in the dialog
 * DiaSetControl(item, value)           puts "value" in "item" in the dialog
 * value = DiaGetControl(item)          gets integer in "item" in the dialog
 * valid = DiaValidEntry(item)          returns nonzero if "item" is nonblank in the dialog
 *
 *    SPECIAL DIALOG CONTROL
 * DiaDimItem(item)                     makes "item" dimmed in the dialog
 * DiaUnDimItem(item)                   makes "item" undimmed in the dialog
 * DiaNoEditControl(item)               makes "item" not editable in the dialog
 * DiaEditControl(item)                 makes "item" editable in the dialog
 * DiaOpaqueEdit(item)                  makes "item" not display text
 * DiaCharacterEdit(item)               makes "item" report keystrokes
 *
 *    CONTROL OF POPUP ITEMS
 * DiaSetPopup(item, count, names)      makes "item" a popup with "count" entries in "names"
 * DiaSetPopupEntry(item, entry)        makes popup "item" use entry "entry" as current
 * entry = DiaGetPopupEntry(item)       gets current entry in popup "item"
 *
 *    CONTROL OF SCROLLABLE ITEMS
 * DiaInitTextDialog(item, top, next, done, sortpos, flags)
 * DiaLoadTextDialog(item, top, next, done, sortpos)
 * DiaStuffLine(item, msg)              add line "msg" to item "item"
 * DiaSelectLine(item, l)               select line "l" in item "item"
 * which = DiaGetCurLine(item)          get current line number in scroll item "item"
 * msg = DiaGetScrollLine(item, l)      get line "l" in scroll item "item"
 * DiaSetScrollLine(item, l, msg)       set line "l" in scroll item "item"
 *
 *    USER-CONTROL OF DIALOG ITEMS
 * DiaItemRect(item, r)                 get rectangle for item "item"
 * DiaPercent(item, p)                  fill item "item" with "p" percent progress
 * DiaRedispRoutine(item, routine)      call "routine" to redisplay item "item"
 * DiaFillPoly(item, x, y, c, r, g, b)  draw "c" points in (x,y) with color (r,g,b)
 * DiaFrameRect(item, r)                draw frame in rectangle "r"
 * DiaDrawLine(item, fx, fy, tx, ty, m) draw line from (fx,fy) to (tx,ty) mode "m"
 * DiaDrawRect(item, r, on)             fill rectangle "r" with "on"
 * DiaGetMouse(x, y)                    get current mouse position into "x/y"
 */
#define	OK           1
#define	CANCEL       2

#define ITEMTYPE   017
#define	BUTTON       0			/* standard push-button */
#define	DEFBUTTON    1			/* default push-button */
#define	CHECK        2			/* standard check-box */
#define	RADIO        3			/* standard radio-button */
#define	SCROLL       4			/* scrolling list of text */
#define	MESSAGE      5			/* non-editable text */
#define	EDITTEXT     6			/* editable text */
#define	ICON         7			/* 32x32 icon */
#define	USERDRAWN  010			/* open area for user graphics */
#define	POPUP      011			/* popup menu */
#define	PROGRESS   012			/* progress bar (0-100%) */
#define	DIVIDELINE 013			/* dividing line */
#define	INACTIVE  0200

/* the last parameter to "DiaInitTextDialog" */
#define	SCSELMOUSE    1			/* mouse clicks select in scroll area */
#define	SCSELKEY      2			/* key strokes select in scroll area */
#define	SCDOUBLEQUIT  4			/* double-click in scroll area exits dialog */
#define	SCREPORT      8			/* report selections in the scroll area */
#define	SCHORIZBAR   16			/* want horizontal scroll-bar in scroll area */
#define	SCSMALLFONT  32			/* want small text in scroll area */
#define	SCFIXEDWIDTH 64			/* want fixed-width text in scroll area */

/* the mode to "DiaDrawLine" */
#define	DLMODEON      1			/* draw the line on */
#define	DLMODEOFF     2			/* draw the line off */
#define	DLMODEINVERT  3			/* draw the line inverted */

typedef struct
{
	INTSML top, left, bottom, right;
} RECTAREA;

typedef struct
{
	INTBIG        data;
	RECTAREA      r;
	unsigned char type;
	char         *msg;
} DIALOGITEM;

typedef struct
{
	RECTAREA    windowRect;
	char       *movable;
	INTSML      items;
	DIALOGITEM *list;
} DIALOG;

#ifdef __cplusplus
extern "C"
{
#endif

/* prototypes for dialog routines */
void   DiaCharacterEdit(INTSML item);
void   DiaDimItem(INTSML item);
void   DiaDoneDialog(void);
void   DiaDrawLine(INTSML item, INTBIG fx, INTBIG fy, INTBIG tx, INTBIG ty, INTBIG mode);
void   DiaDrawRect(INTSML item, RECTAREA *r, INTSML on);
void   DiaEditControl(INTSML item);
void   DiaFillPoly(INTSML item, INTBIG *x, INTBIG *y, INTBIG count, INTBIG r, INTBIG g, INTBIG b);
void   DiaFrameRect(INTSML item, RECTAREA *r);
INTSML DiaGetControl(INTSML item);
INTSML DiaGetCurLine(INTSML item);
void   DiaGetMouse(INTSML *x, INTSML *y);
INTSML DiaGetNextCharacter(INTSML *item);
INTSML DiaGetPopupEntry(INTSML item);
char  *DiaGetScrollLine(INTSML item, INTSML which);
char  *DiaGetText(INTSML item);
INTSML DiaInitDialog(DIALOG *dialog);
void   DiaInitTextDialog(INTSML item, INTSML (*toplist)(char **),
		char *(*nextinlist)(void), void (*donelist)(void),
		INTSML sortpos, INTSML flags);
void   DiaItemRect(INTSML item, RECTAREA *rect);
void   DiaLoadTextDialog(INTSML item, INTSML (*toplist)(char **), char *(*nextinlist)(void),
		void (*donelist)(void), INTSML sortpos);
INTSML DiaNextHit(void);
void   DiaNoEditControl(INTSML item);
void   DiaOpaqueEdit(INTSML item);
void   DiaPercent(INTSML item, INTBIG p);
void   DiaRedispRoutine(INTSML item, void (*routine)(RECTAREA*));
void   DiaSelectLine(INTSML item, INTSML which);
void   DiaSetControl(INTSML item, INTSML value);
void   DiaSetPopup(INTSML item, INTSML count, char **names);
void   DiaSetPopupEntry(INTSML item, INTSML entry);
void   DiaSetScrollLine(INTSML item, INTSML which, char *msg);
void   DiaSetText(INTSML item, char *msg);
void   DiaStuffLine(INTSML item, char *line);
void   DiaUnDimItem(INTSML item);
INTSML DiaValidEntry(INTSML item);

#ifdef __cplusplus
}
#endif
