/*
 * Electric(tm) VLSI Design System
 *
 * File: ioedifi.c
 * Input/output aid: EDIF 2 0 0 netlist reader
 * Written by: Glen Lawson
 *
 * Copyright (c) 1999 Electric Editor Incorporated.
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 *
 * Electric Editor Incorporated
 * 23470 Sunset Drive, Suite 108
 * Los Gatos, California 95033
 * support@electriceditor.com
 */

/*
 * Notes (11/12/92):
 *	I have tried EDIF files from CADENCE and VALID only.
 *	I used splines for the arcs, should really use a geometric arc primitive
 *	No justification of text (text is there, but needs improvement)
 *	Does not fully support portbundles
 *	Multiple ports of the same name are named port_x (x is 1 to n duplicate)
 *	Keywords such as ARRAY have unnamed parameters, ie (array (name..) 5 6)
 *	this is handled in the io_edprocess_integer function called by io_edget_keyword,
 *	this is a kludge to fix this problem, a real table driven parser should be used.
 * Improved (3/9/92) by Steven Rubin:
 *	handle progress dialogs during input
 *	changed "malloc" and "free" to "emalloc" and "efree"
 *      memory allocation bug in "io_edpop_stack()" caused crash when a technology with GDS
 *         layers is current
 * gml (1/94)
 *	Use circle arcs instead of splines.
 *	Support text justifications and text height
 * 	Better NAME/RENAME/STRINGDISPLAY/ANNOTATE text handling.
 * gml (3/94)
 *  	ANSI prototypes
 * gml (9/94)
 *	Changed arcs to simple polygons plus ARC attribute
 * gml (12/94)
 *      Can read NETLIST views
 */

#include "config.h"
#if IOEDIF

#include "global.h"
#include "egraphics.h"
#include "efunction.h"
#include "database.h"
#include "edialogs.h"
#include "tecgen.h"
#include "tecart.h"
#include "tecschem.h"
#include "eio.h"
#include "usr.h"
#include <math.h>

/* #define ED_FACETCENTER 1 */  /* uncomment to create facet center markers */

/* for the "qsort" function */
typedef int (*__cmp_func)(const void *, const void *);

/* the default cell name */
#define CELLNAME "EDIF"

/* some length defines */
#define LINE 1024
#define WORD 512
#define MAXLAYERS 256

/* generic defines */
#define TRUE 1
#define FALSE 0
#define WHITE_SPACE " \t\r\n"
static char *DELIMETERS =  " \t\r\n=.:;'()}\"";
#define SEMI ';'
#define COLON ':'
#define TICK '\047'
#define PERIOD '.'
#define LEFTP '('
#define RIGHTP ')'
#define LEFTC '{'
#define RIGHTC '}'
#define LEFTB '['
#define RIGHTB ']'
#define EQUAL '='
#define QUOTE '\"'
#define SPACE ' '
#define TAB '\t'
#define NEWLINE '\n'
#define RETURN '\r'
#define SPECIAL_CHARS "-^*#"

typedef enum _textjust
{
	UPPERLEFT,
	UPPERCENTER,
	UPPERRIGHT,
	CENTERLEFT,
	CENTERCENTER,
	CENTERRIGHT,
	LOWERLEFT,
	LOWERCENTER,
	LOWERRIGHT
} TEXTJUST;

/* name table for layers ... */
typedef struct _nametable
{
	char *original;					/* the original MASK layer */
	char *replace;					/* the replacement layer */
	NODEPROTO *node;				/* the basic electric node */
	ARCPROTO *arc;					/* the basic arc type */
	int textheight;					/* default text height */
	enum _textjust justification;	/* default justification */
	int visible; 					/* layer is visible */
} NAMETABLE, *NAMETABLE_PTR;
#define NONAMETABLE ( ( NAMETABLE_PTR ) 0 )

/* the following is a table of EDIF keywords, and functions */
typedef enum _kstate
{
	KUNKNOWN,
	KINIT,
	KANNOTATE,
	KARC,
	KARRAY,
	KAUTHOR,
	KBOOLEAN,
	KBORDERWIDTH,
	KBOUNDINGBOX,
	KCELL,
	KCELLREF,
	KCELLTYPE,
	KCIRCLE,
	KCOLOR,
	KCONTENTS,
	KCOMMENT,
	KCOMMENTGRAPHICS,
	KCONNECTLOCATION,
	KCORNERTYPE,
	KCURVE,
	KDATAORIGIN,
	KDCFANOUTLOAD,
	KDCMAXFANOUT,
	KDELTA,
	KDESIGN,
	KDESIGNATOR,
	KDIRECTION,
	KDISPLAY,
	KDOT,
	KEDIF,
	KEDIFLEVEL,
	KEDIFVERSION,
	KENDTYPE,
	KEXTERNAL,
	KFABRICATE,
	KFALSE,
	KFIGURE,
	KFIGUREGROUP,
	KFIGUREGROUPOVERRIDE,
	KFIGUREGROUPREF,
	KFILLPATTERN,
	KGRIDMAP,
	KINSTANCE,
	KINSTANCEREF,
	KINTEGER,
	KINTERFACE,
	KJOINED,
	KJUSTIFY,
	KKEYWORDDISPLAY,
	KEDIFKEYMAP,
	KEDIFKEYLEVEL,
	KLIBRARY,
	KLIBRARYREF,
	KLISTOFNETS,
	KLISTOFPORTS,
	KLOGICREF,
	KMEMBER,
	KMUSTJOIN,
	KNAME,
	KNET,
	KNETBUNDLE,
	KNETREF,
	KNUMBER,
	KNUMBERDEFINITION,
	KORIENTATION,
	KORIGIN,
	KOWNER,
	KPAGE,
	KPAGESIZE,
	KPARAMETER,
	KPATH,
	KPATHWIDTH,
	KPOINT,
	KPOINTLIST,
	KPOLYGON,
	KPORT,
	KPORTBUNDLE,
	KPORTIMPLEMENTATION,
	KPORTINSTANCE,
	KPORTLIST,
	KPORTREF,
	KPROGRAM,
	KPROPERTY,
	KPROPERTYDISPLAY,
	KPT,
	KNAMEDEF,
	KOPENSHAPE,
	KRECTANGLE,
	KRENAME,
	KSCALE,
	KSCALEDINTEGER,
	KSCALEX,
	KSCALEY,
	KSITE,
	KSHAPE,
	KSTATUS,
	KSTRING,
	KSTRINGDISPLAY,
	KSYMBOL,
	KTECHNOLOGY,
	KTEXTHEIGHT,
	KTIMESTAMP,
	KTRANSFORM,
	KTRUE,
	KUNIT,
	KUSERDATA,
	KVERSION,
	KVIEW,
	KVIEWLIST,
	KVIEWREF,
	KVIEWTYPE,
	KVISIBLE,
	KWEAKJOINED,
	KWRITTEN
} KSTATE;

/* forware declaration */
struct _edifgbl;

#define KNULL ( ( enum _kstate * ) 0 )
typedef struct _edifkey
{
	char *name;							/* the name of the keyword */
	INTBIG (*func)(struct _edifgbl*);	/* the function to execute */
	enum _kstate next;					/* the next state, if any */
	enum _kstate *state;				/* edif state */
} EDIFKEY, *EDIFKEY_PTR;
#define NOEDIFKEY ( ( EDIFKEY_PTR ) 0 )

/* Edif viewtypes ... */

typedef enum _vtypes
{
	VNULL,
	VBEHAVIOR,
	VDOCUMENT,
	VGRAPHIC,
	VLOGICMODEL,
	VMASKLAYOUT,
	VNETLIST,
	VPCBLAYOUT,
	VSCHEMATIC,
	VSTRANGER,
	VSYMBOLIC,
	VSYMBOL 	/* not a real EDIF view, but electric has one */
} VTYPES;

/* Edif geometry types ... */
typedef enum _gtypes
{
	GUNKNOWN,
	GRECTANGLE,
	GPOLYGON,
	GSHAPE,
	GOPENSHAPE,
	GTEXT,
	GPATH,
	GINSTANCE,
	GCIRCLE,
	GARC,
	GPIN,
	GNET,
	GBUS
} GTYPES;

/* 8 standard orientations */
typedef enum _otypes
{
	OUNKNOWN,
	OR0,
	OR90,
	OR180,
	OR270,
	OMX,
	OMY,
	OMYR90,
	OMXR90
} OTYPES;

typedef struct _ept
{
	int x , y;
	struct _ept *nextpt;
} EPT, *EPT_PTR;
#define NOEPT ( ( EPT_PTR ) 0 )

typedef enum _edportdir { INOUT, INPUT, OUTPUT } EDPORTDIR;

#define INCH (10*edifgbl->technology->deflambda)

typedef struct _edport
{
	char *name;
	char *reference;
	enum _edportdir direction;
	int arrayx, arrayy;
	struct _edport *next;
} EDPORT, *EDPORT_PTR;
#define NOEDPORT ( ( EDPORT_PTR ) 0 )

typedef struct _ednetport
{
	NODEINST *ni;					/* unique node port is attached to */
	PORTPROTO *pp;					/* port type (no portarc yet) */
	int member;						/* member for an array */
	struct _ednetport *next;		/* next common to a net */
} EDNETPORT, *EDNETPORT_PTR;
#define NOEDNETPORT ( ( EDNETPORT_PTR ) 0 )

typedef enum _edvendor
{
	EVUNKNOWN,
	EVCADENCE,
	EVVALID,
	EVSYNOPSYS,
	EVMENTOR,
	EVVIEWLOGIC
} EDVENDOR;

#define MAXSHEETS 6
typedef enum _sheet_size
{
	SHEET_ASIZE = 0,
	SHEET_BSIZE = 1,
	SHEET_CSIZE = 2,
	SHEET_DSIZE = 3,
	SHEET_ESIZE = 4,
	SHEET_INFSIZE = 5
} SHEET_SIZE;

typedef enum _property_type
{
	PUNKNOWN,
	PSTRING,
	PINTEGER,
	PNUMBER
} PROPERTY_TYPE;

typedef struct _edproperty
{
	char *name;
	union
	{
		int integer;
		float number;
		char *string;
	} val;
	enum _property_type type;
	struct _edproperty *next;
} EDPROPERTY, *EDPROPERTY_PTR;
#define NOEDPROPERTY ( ( EDPROPERTY_PTR ) 0 )

/* edifgbl declaration */
typedef struct _edifgbl
{
	/* general information variables */
	FILE *edif_file;             	/* the input file */
	jmp_buf env;					/* longjmp return */

	/* view information ... */
	int pageno;						/* the schematic page number */
	enum _vtypes  active_view;		/* indicates we are in a NETLIST view */
	int drawing_number;				/* the current drawing number */
	int p_ref, c_ref;				/* peripheral and core reference names */
	enum _edvendor vendor;			/* the current vendor type */

	/* parser variables ... */
	enum _kstate state;          	/* the current parser state */
	int lineno;                  	/* the current line number */
	char buffer[LINE+1];         	/* the read buffer */
	char *pos;                   	/* the position within the buffer */
	int ignoreblock;				/* no update flag */
	char *delimeters;				/* parsing delimeters for edif */
	int errors, warnings;			/* load status */

	/* electric context data ... */
	LIBRARY *library;				/* the new library */
	TECHNOLOGY *technology;			/* the active technology */
	NODEPROTO *current_cell;		/* the current active cell */
	NODEINST *current_node;			/* the current active instance */
	ARCINST *current_arc;			/* the current active arc */
	NODEPROTO *figure_group;		/* the current figure group node */
	ARCPROTO *arc_type;				/* the current (if exists) arc type */
	NODEPROTO *proto;				/* the cellRef type */
	PORTPROTO *current_port;		/* the current port proto */

	/* general geometry information ... */
	enum _gtypes geometry;			/* the current geometry type */
	struct _ept *points;			/* the list of points */
	struct _ept *lastpt;			/* the end of the list */
	int pt_count;					/* the count of points */
	enum _otypes orientation;		/* the orientation of the structure */
	enum _edportdir direction;		/* port direction */

	/* geometric path constructors ... */
	int path_width;					/* the width of the path */
	int extend_corner;				/* extend path corner flag */
	int extend_end;              	/* extend path end flag */

	/* array variables ... */
	int isarray;					/* set if truely an array */
	int arrayx, arrayy;				/* the bounds of the array */
	int deltaxX, deltaxY;			/* offsets in x and y for an X increment */
	int deltayX, deltayY;			/* offsets in x and y for an Y increment */
	int deltapts;					/* which offset flag */
	int memberx, membery;			/* the element of the array */

	/* text variables ... */
	char string[LINE+1];			/* Text string */
	int textheight;					/* the current default text height */
	enum _textjust justification;	/* the current text justificaton */
	int visible;					/* is stringdisplay visible */
	struct {						/* save block for name and rename strings */
		char string[LINE+1];		/* the saved string, if NULL not saved */
		struct _ept *points;		/* origin x and y */
		int pt_count;				/* count of points (usually 1) */
		int textheight;				/* the height of text */
		int visible;				/* visiblity of the text */
		enum _textjust justification; 	/* justification of the text */
		enum _otypes orientation; 	/* orientation of the text */
	} save_text;

	/* technology data ... */
	double scale;					/* scaling value */
	double dbunit;					/* database units */
	double meters;					/* per number of meters */
	double val1, val2;				/* temporary storage */

	/* current name and rename of EDIF objects ... */
	char cell_reference[WORD+1];	/* the current cell name */
	char cell_name[WORD+1];			/* the current cell name (original) */
	char port_reference[WORD+1];	/* the current port name */
	char port_name[WORD+1];			/* the current port name (original) */
	char instance_reference[WORD+1]; /* the current instance name */
	char instance_name[WORD+1];		/* the current instance name (original) */
	char bundle_reference[WORD+1];	/* the current bundle name */
	char bundle_name[WORD+1];		/* the current bundle name (original) */
	char net_reference[WORD+1];		/* the current net name */
	char net_name[WORD+1];			/* the current net name (original) */
	char property_reference[WORD+1]; /* the current property name */
	char property_name[WORD+1];		/* the current property name (original) */

	enum _property_type ptype;		/* the type of property */
	union {
		char *string;				/* string buffer */
		int integer;				/* integer buffer */
		float number;				/* number buffer */
	} pval;

	char name[WORD+1];				/* the name of the object */
	char original[WORD+1];			/* the original name of the object */

	/* layer or figure information ... */
	int layer_ptr;					/* pointer to layer table entry */
	NAMETABLE_PTR nametbl[MAXLAYERS]; /* the current name mapping table */
	NAMETABLE_PTR cur_nametbl; 		/* the current name table entry */

	/* cell name lookup (from rename) */
	NAMETABLE_PTR *celltbl;			/* the cell lookup table */
	int celltbl_cnt;				/* the count of entries in the cell table */
	int celltbl_sze;				/* maximum size of the table */

	/* port data for port exporting */
	EDPORT_PTR ports;				/* active port list */
	EDPORT_PTR free_ports;			/* the list of free ports */

	/* property data for all objects */
	EDPROPERTY_PTR properties;		/* active property list */
	EDPROPERTY_PTR free_properties;	/* the list of free properties */

	/* net constructors */
	EDNETPORT_PTR firstnetport;		/* the list of ports on a net */
	EDNETPORT_PTR lastnetport; 		/* the last in the list on a net */
	EDNETPORT_PTR free_netports;	/* the list of free ports */

	/* view NETLIST layout */
	enum _sheet_size sheet_size;	/* the size of the target sheet */
	int sh_xpos, sh_ypos;			/* current sheet position */
	int sh_offset;					/* next x offset */
	int ipos, bpos, opos;			/* current position pointers */
} EDIFGBL, *EDIFGBL_PTR;
#define NOEDIFGBL ( ( EDIFGBL_PTR ) -1 )

INTBIG io_edload_edif(EDIFGBL_PTR edifgbl, INTBIG key_count, EDIFKEY_PTR keywords);
INTBIG io_edfreesavedptlist(EDIFGBL_PTR edifgbl);
double io_edgetnumber(EDIFGBL_PTR edifgbl);
INTBIG io_edcheck_name(EDIFGBL_PTR edifgbl);
INTBIG io_edpop_stack(EDIFGBL_PTR edifgbl);
void io_edprocess_integer(EDIFGBL_PTR edifgbl, INTBIG value);
INTSML io_edgetrot(OTYPES orientation);
INTSML io_edgettrans(OTYPES orientation);
INTBIG io_edclockwise(EPT_PTR pt1, EPT_PTR pt2, EPT_PTR pt3);
INTBIG io_edcompare_name(NAMETABLE_PTR *name1, NAMETABLE_PTR *name2);
INTBIG io_edis_integer(char *buffer);
char *io_edget_keyword(EDIFGBL_PTR edifgbl, char *buffer);
char *io_edget_token(EDIFGBL_PTR edifgbl, char *buffer, char idelim);
char *io_edpos_token(EDIFGBL_PTR edifgbl);
char *io_edflush_line(EDIFGBL_PTR edifgbl, char delim);
INTSML io_edget_line(char *line, INTSML limit, FILE *file);
void io_edget_delim(EDIFGBL_PTR edifgbl, char delim);
NODEINST *io_edreplacenodeinst(NODEINST *ni, NODEPROTO *np, PORTPROTO *pp, INTBIG x, INTBIG y);
char *io_edstralloc(char *str);
INTBIG io_edallocport(EDIFGBL_PTR edifgbl);
INTBIG io_edallocproperty(EDIFGBL_PTR edifgbl, char *name, PROPERTY_TYPE type, int integer,
	float number, char *string);
INTBIG io_edallocnetport(EDIFGBL_PTR edifgbl);
void io_edeq_of_a_line(double sx, double sy, double ex, double ey, double *A, double *B, double *C);
INTBIG io_eddetermine_intersection(double A[2], double B[2], double C[2], double *x, double *y);
void io_ededif_arc(INTBIG x[3], INTBIG y[3], INTBIG *xc, INTBIG *yc, INTBIG *r, double *so,
	double *ar, INTBIG *rot, INTBIG *trans);
NODEPROTO *io_edmakeiconfacet(PORTPROTO *fpp, char *iconname, char *pt,
TECHNOLOGY *tech, LIBRARY *lib);
void io_ednamearc(EDIFGBL *edifgbl, ARCINST *ai);
INTBIG io_edfindport(NODEPROTO *facet, INTBIG x, INTBIG y, ARCPROTO *ap, NODEINST *nis[],
	PORTPROTO *pps[]);
INTBIG io_edfindport_geom(NODEPROTO *facet, INTBIG x, INTBIG y, ARCPROTO *ap,
	NODEINST *nis[], PORTPROTO *pps[], ARCINST **ai);

/* keyword parser entry points, only these keywords are supported by this parser,
   if you need more, add them. Each entry point can have an exit point in the
   function io_edpop_stack. This is processed when ')' is encountered in the EDIF
   file.
 */
static INTBIG io_ednoop(EDIFGBL_PTR edifgbl);
static INTBIG io_edtechnology(EDIFGBL_PTR edifgbl);
static INTBIG io_edfabricate(EDIFGBL_PTR edifgbl);
static INTBIG io_edpage(EDIFGBL_PTR edifgbl);
static INTBIG io_edfigure(EDIFGBL_PTR edifgbl);
static INTBIG io_edfigureGroupOverride(EDIFGBL_PTR edifgbl);
static INTBIG io_edscale(EDIFGBL_PTR edifgbl);
static INTBIG io_edendType(EDIFGBL_PTR edifgbl);
static INTBIG io_edcornerType(EDIFGBL_PTR edifgbl);
static INTBIG io_edstringDisplay(EDIFGBL_PTR edifgbl);
static INTBIG io_edtrue(EDIFGBL_PTR edifgbl);
static INTBIG io_edfalse(EDIFGBL_PTR edifgbl);
static INTBIG io_edgetlayer(EDIFGBL_PTR edifgbl);
static INTBIG io_edpointList(EDIFGBL_PTR edifgbl);
static INTBIG io_edpoint(EDIFGBL_PTR edifgbl);
static INTBIG io_edjustify(EDIFGBL_PTR edifgbl);
static INTBIG io_edtextHeight(EDIFGBL_PTR edifgbl);
static INTBIG io_edpt(EDIFGBL_PTR edifgbl);
static INTBIG io_edorientation(EDIFGBL_PTR edifgbl);
static INTBIG io_edpathWidth(EDIFGBL_PTR edifgbl);
static INTBIG io_edrectangle(EDIFGBL_PTR edifgbl);
static INTBIG io_edcircle(EDIFGBL_PTR edifgbl);
static INTBIG io_edpath(EDIFGBL_PTR edifgbl);
static INTBIG io_edpolygon(EDIFGBL_PTR edifgbl);
static INTBIG io_edopenShape(EDIFGBL_PTR edifgbl);
static INTBIG io_edfreeptlist(EDIFGBL_PTR edifgbl);
static INTBIG io_edlibrary(EDIFGBL_PTR edifgbl);
static INTBIG io_eddesign(EDIFGBL_PTR edifgbl);
static INTBIG io_edcell(EDIFGBL_PTR edifgbl);
static INTBIG io_edview(EDIFGBL_PTR edifgbl);
static INTBIG io_edviewType(EDIFGBL_PTR edifgbl);
static INTBIG io_edcontents(EDIFGBL_PTR edifgbl);
static INTBIG io_edport(EDIFGBL_PTR edifgbl);
static INTBIG io_eddirection(EDIFGBL_PTR edifgbl);
static INTBIG io_edinstance(EDIFGBL_PTR edifgbl);
static INTBIG io_edarray(EDIFGBL_PTR edifgbl);
static INTBIG io_eddelta(EDIFGBL_PTR edifgbl);
static INTBIG io_edcellRef(EDIFGBL_PTR edifgbl);
static INTBIG io_edproperty(EDIFGBL_PTR edifgbl);
static INTBIG io_ednet(EDIFGBL_PTR edifgbl);
static INTBIG io_ednetBundle(EDIFGBL_PTR edifgbl);
static INTBIG io_edjoined(EDIFGBL_PTR edifgbl);
static INTBIG io_edportRef(EDIFGBL_PTR edifgbl);
static INTBIG io_edinstanceRef(EDIFGBL_PTR edifgbl);
static INTBIG io_edmember(EDIFGBL_PTR edifgbl);
static INTBIG io_edname(EDIFGBL_PTR edifgbl);
static INTBIG io_edrename(EDIFGBL_PTR edifgbl);
static INTBIG io_edboundingBox(EDIFGBL_PTR edifgbl);
static INTBIG io_edsymbol(EDIFGBL_PTR edifgbl);
static INTBIG io_edportImplementation(EDIFGBL_PTR edifgbl);
static INTBIG io_edconnectLocation(EDIFGBL_PTR edifgbl);
static INTBIG io_eddot(EDIFGBL_PTR edifgbl);
static INTBIG io_edprogram(EDIFGBL_PTR edifgbl);
static INTBIG io_edunit(EDIFGBL_PTR edifgbl);
static INTBIG io_edinteger(EDIFGBL_PTR edifgbl);
static INTBIG io_ednumber(EDIFGBL_PTR edifgbl);
static INTBIG io_edstring(EDIFGBL_PTR edifgbl);
static INTBIG io_edinterface(EDIFGBL_PTR edifgbl);

/* Parser keyword state tables, only some are provided, others are simply incomplete.
   If you have time, please feel free to complete these keyword tables and the following
   parse array. Note that a bad state does not cause the parser to quit, the action is
   still executed.
 */
static enum _kstate Sarray[] = {KINSTANCE, KPORT, KNET, KUNKNOWN};
static enum _kstate Sauthor[] = {KWRITTEN, KUNKNOWN};
static enum _kstate SboundingBox[] = {KSYMBOL, KCONTENTS, KUNKNOWN};
static enum _kstate Scell[] = {KEXTERNAL, KLIBRARY, KUNKNOWN};
static enum _kstate ScellRef[] = {KDESIGN, KVIEWREF, KINSTANCE, KUNKNOWN};
static enum _kstate ScellType[] = {KCELL, KUNKNOWN};
static enum _kstate Scontents[] = {KVIEW, KUNKNOWN};
static enum _kstate Sdesign[] = {KEDIF, KUNKNOWN};
static enum _kstate Sdirection[] = {KPORT, KUNKNOWN};
static enum _kstate Sedif[] = {KINIT, KUNKNOWN};
static enum _kstate SedifLevel[] = {KEDIF, KEXTERNAL, KLIBRARY, KUNKNOWN};
static enum _kstate SedifVersion[] = {KEDIF, KUNKNOWN};
static enum _kstate Sinstance[] = {KCONTENTS, KPAGE, KPORTIMPLEMENTATION, KCOMMENTGRAPHICS, KUNKNOWN};
static enum _kstate SinstanceRef[] = {KINSTANCEREF, KPORTREF, KUNKNOWN};
static enum _kstate Sinterface[] = {KVIEW, KUNKNOWN};
static enum _kstate Sjoined[] = {KINTERFACE, KNET, KMUSTJOIN, KWEAKJOINED, KUNKNOWN};
static enum _kstate SkeywordMap[] = {KEDIF, KUNKNOWN};
static enum _kstate Slibrary[] = {KEDIF, KUNKNOWN};
static enum _kstate SlibraryRef[] = {KCELLREF, KFIGUREGROUPREF, KLOGICREF, KUNKNOWN};
static enum _kstate SlistOfNets[] = {KNETBUNDLE, KUNKNOWN};
static enum _kstate Snet[] = {KCONTENTS, KPAGE, KLISTOFNETS, KUNKNOWN};
static enum _kstate SnetBundle[] = {KCONTENTS, KPAGE, KUNKNOWN};
static enum _kstate SnumberDefinition[] = {KTECHNOLOGY, KUNKNOWN};
static enum _kstate Sport[] = {KINTERFACE, KLISTOFPORTS, KUNKNOWN};
static enum _kstate Sscale[] = {KNUMBERDEFINITION, KUNKNOWN};
static enum _kstate Sstatus[] = {KCELL, KDESIGN, KEDIF, KEXTERNAL, KLIBRARY, KVIEW, KUNKNOWN};
static enum _kstate Ssymbol[] = {KINTERFACE, KUNKNOWN};
static enum _kstate Stechnology[] = {KEXTERNAL, KLIBRARY, KUNKNOWN};
static enum _kstate Stimestamp[] = {KWRITTEN, KUNKNOWN};
static enum _kstate Sunit[] = {KPARAMETER, KPROPERTY, KSCALE, KUNKNOWN};
static enum _kstate Sview[] = {KCELL, KUNKNOWN};
static enum _kstate SviewRef[] = {KINSTANCE, KINSTANCEREF, KNETREF, KPORTREF, KSITE, KVIEWLIST, KUNKNOWN};
static enum _kstate SviewType[] = {KVIEW, KUNKNOWN};
static enum _kstate Swritten[] = {KSTATUS, KUNKNOWN};

/* To add a new keyword, insert into this sorted list and into the enumerated keyword
   list in ioedif.h. If no action is required, just use the io_ednoop as the action
   routine. If some action is required, add a new function. Note that all operations
   are dependent on the current state. If a keyword is correct, but the required state
   table is not, add it to the above data structure (_kstate). Or make it non-state
   dependent with a KNULL in the last field.
 */
static EDIFKEY edif_keywords[] =
{
	{"annotate", io_ednoop, KANNOTATE, KNULL},
	{"arc", io_ednoop, KARC, KNULL},
	{"array", io_edarray, KARRAY, Sarray},
	{"author", io_ednoop, KAUTHOR, Sauthor},
	{"boolean", io_ednoop, KBOOLEAN, KNULL},
	{"borderWidth", io_ednoop, KBORDERWIDTH, KNULL},
	{"boundingBox", io_edboundingBox, KBOUNDINGBOX, SboundingBox},
	{"cell", io_edcell, KCELL, Scell },
	{"cellRef", io_edcellRef, KCELLREF, ScellRef},
	{"cellType", io_ednoop, KCELLTYPE, ScellType},
	{"circle", io_edcircle, KCIRCLE, KNULL},
	{"color", io_ednoop, KCOLOR, KNULL},
	{"comment", io_ednoop, KCOMMENT, KNULL},
	{"commentGraphics", io_ednoop, KCOMMENTGRAPHICS, KNULL},
	{"connectLocation", io_edconnectLocation, KCONNECTLOCATION, KNULL},
	{"contents", io_edcontents, KCONTENTS, Scontents},
	{"cornerType", io_edcornerType, KCORNERTYPE, KNULL},
	{"curve", io_ednoop, KCURVE, KNULL},
	{"dataOrigin", io_ednoop, KDATAORIGIN, KNULL},
	{"dcFanoutLoad", io_ednoop, KDCFANOUTLOAD, KNULL},
	{"dcMaxFanout", io_ednoop, KDCMAXFANOUT, KNULL},
	{"delta", io_eddelta, KDELTA, KNULL},
	{"design", io_eddesign, KDESIGN, Sdesign},
	{"designator", io_ednoop, KDESIGNATOR, KNULL},
	{"direction", io_eddirection, KDIRECTION, Sdirection},
	{"display", io_edfigure, KDISPLAY, KNULL},
	{"dot", io_eddot, KDOT, KNULL},
	{"e", io_ednoop, KSCALEDINTEGER, KNULL},
	{"edif", io_ednoop, KEDIF, Sedif},
	{"edifLevel", io_ednoop, KEDIFLEVEL, SedifLevel},
	{"edifVersion", io_ednoop, KEDIFVERSION, SedifVersion},
	{"endType", io_edendType, KENDTYPE, KNULL},
	{"external", io_edlibrary, KEXTERNAL, Slibrary},
	{"fabricate", io_edfabricate, KFABRICATE, KNULL},
	{"false", io_edfalse, KFALSE, KNULL},
	{"figure", io_edfigure, KFIGURE, KNULL},
	{"figureGroup", io_ednoop, KFIGUREGROUP, KNULL},
	{"figureGroupOverride", io_edfigureGroupOverride, KFIGUREGROUPOVERRIDE, KNULL},
	{"fillpattern", io_ednoop, KFILLPATTERN, KNULL},
	{"gridMap", io_ednoop, KGRIDMAP, KNULL},
	{"instance", io_edinstance, KINSTANCE, Sinstance},
	{"instanceRef", io_edinstanceRef, KINSTANCEREF, SinstanceRef},
	{"integer", io_edinteger, KINTEGER, KNULL},
	{"interface", io_edinterface, KINTERFACE, Sinterface},
	{"joined", io_edjoined, KJOINED, Sjoined},
	{"justify", io_edjustify, KJUSTIFY, KNULL},
	{"keywordDisplay", io_ednoop, KKEYWORDDISPLAY, KNULL},
	{"keywordLevel", io_ednoop, KEDIFKEYLEVEL, KNULL},
	{"keywordMap", io_ednoop, KEDIFKEYMAP, SkeywordMap},
	{"library", io_edlibrary, KLIBRARY, Slibrary},
	{"libraryRef", io_ednoop, KLIBRARYREF, SlibraryRef},
	{"listOfNets", io_ednoop, KLISTOFNETS, SlistOfNets},
	{"listOfPorts", io_ednoop, KLISTOFPORTS, KNULL},
	{"member", io_edmember, KMEMBER, KNULL},
	{"name", io_edname, KNAME, KNULL},
	{"net", io_ednet, KNET, Snet},
	{"netBundle", io_ednetBundle, KNETBUNDLE, SnetBundle},
	{"number", io_ednumber, KNUMBER, KNULL},
	{"numberDefinition", io_ednoop, KNUMBERDEFINITION, SnumberDefinition},
	{"openShape", io_edopenShape, KOPENSHAPE, KNULL},
	{"orientation", io_edorientation, KORIENTATION, KNULL},
	{"origin", io_ednoop, KORIGIN, KNULL},
	{"owner", io_ednoop, KOWNER, KNULL},
	{"page", io_edpage, KPAGE, KNULL},
	{"pageSize", io_ednoop, KPAGESIZE, KNULL},
	{"path", io_edpath, KPATH, KNULL},
	{"pathWidth", io_edpathWidth, KPATHWIDTH, KNULL},
	{"point", io_edpoint, KPOINT, KNULL},
	{"pointList", io_edpointList, KPOINTLIST, KNULL},
	{"polygon", io_edpolygon, KPOLYGON, KNULL},
	{"port", io_edport, KPORT, Sport},
	{"portBundle", io_ednoop, KPORTBUNDLE, KNULL},
	{"portImplementation", io_edportImplementation, KPORTIMPLEMENTATION, KNULL},
	{"portInstance", io_ednoop,KPORTINSTANCE, KNULL},
	{"portList", io_ednoop, KPORTLIST, KNULL},
	{"portRef", io_edportRef, KPORTREF, KNULL},
	{"program", io_edprogram, KPROGRAM, KNULL},
	{"property", io_edproperty, KPROPERTY, KNULL},
	{"propertyDisplay", io_ednoop, KPROPERTYDISPLAY, KNULL},
	{"pt", io_edpt, KPT, KNULL},
	{"rectangle", io_edrectangle, KRECTANGLE, KNULL},
	{"rename", io_edrename, KRENAME, KNULL},
	{"scale", io_edscale, KSCALE, Sscale},
	{"scaleX", io_ednoop, KSCALEX, KNULL},
	{"scaleY", io_ednoop, KSCALEY, KNULL},
	{"shape", io_ednoop, KSHAPE, KNULL},
	{"status", io_ednoop, KSTATUS, Sstatus},
	{"string", io_edstring, KSTRING, KNULL},
	{"stringDisplay", io_edstringDisplay, KSTRINGDISPLAY, KNULL},
	{"symbol", io_edsymbol, KSYMBOL, Ssymbol},
	{"technology", io_edtechnology, KTECHNOLOGY, Stechnology},
	{"textHeight", io_edtextHeight, KTEXTHEIGHT, KNULL},
	{"timestamp", io_ednoop, KTIMESTAMP, Stimestamp},
	{"transform", io_ednoop, KTRANSFORM, KNULL},
	{"true", io_edtrue, KTRUE, KNULL},
	{"unit", io_edunit, KUNIT, Sunit},
	{"userData", io_ednoop, KUSERDATA, KNULL},
	{"version", io_ednoop, KVERSION, KNULL},
	{"view", io_edview, KVIEW, Sview},
	{"viewRef", io_ednoop, KVIEWREF, SviewRef},
	{"viewType", io_edviewType, KVIEWTYPE, SviewType},
	{"visible", io_ednoop, KVISIBLE, KNULL},
	{"written", io_ednoop, KWRITTEN, Swritten},
	0};

/* edif keyword stack, this stack matches the current depth of the EDIF file,
   for the keyword instance it would be KEDIF, KLIBRARY, KCELL, KCONTENTS, KINSTANCE
 */
static KSTATE kstack[1000];
static INTBIG kstack_ptr = 0;
static INTBIG filelen, filepos;

/* some standard artwork primitivies */
static PORTPROTO *default_port = NOPORTPROTO;
static PORTPROTO *default_iconport = NOPORTPROTO;
static PORTPROTO *default_busport = NOPORTPROTO;
static PORTPROTO *default_input = NOPORTPROTO;
static PORTPROTO *default_output = NOPORTPROTO;
INTBIG EDIF_name_key = 0;
static INTBIG EDIF_array_key = 0;
static INTBIG EDIF_annotate_key = 0;
static struct
{
	int width;
	int height;
} EDIF_sheet_bounds[MAXSHEETS] =
{
	{8, 10},
	{16, 10},
	{16, 20},
	{32, 20},
	{32, 40},
	{64, 40}
};

#define RET_NOMEMORY() { ttyputerr("error: no memory"); return 1; }

/*
 * routine to write a ".edif" file from the library "lib"
 */
INTSML io_readediflibrary(LIBRARY *lib)
{
	EDIFGBL edifgbl;
	INTBIG key_count, i;
	char *filename, *msg;
	NODEPROTO *np;
	extern DIALOG us_progressdialog;
	void net_examinenodeproto(NODEPROTO *np);

	if ((edifgbl.edif_file = xopen(lib->libfile, FILETYPEEDIF, "", &filename)) == NULL)
	{
		ttyputerr("File %s not found", lib->libfile);
		return(1);
	}
	filelen = filesize(edifgbl.edif_file);
	filepos = 0;

	/* parser inits */
	edifgbl.state = KINIT;
	edifgbl.lineno = 1;
	edifgbl.buffer[0] = '\0';
	edifgbl.pos = edifgbl.buffer;
	edifgbl.delimeters = DELIMETERS;
	edifgbl.errors = edifgbl.warnings = 0;
	edifgbl.ignoreblock = 0;
	edifgbl.vendor = EVUNKNOWN;

	/* general inits */
	edifgbl.scale = 1.0;
	edifgbl.dbunit = 1.0;
	/*  edifgbl.meters = 0.000000001; */ /* default 1 db unit per milli-micron */
	edifgbl.meters = scaletodispunit(1, DISPUNITCM) / 100.0;
	edifgbl.library = lib;
	edifgbl.technology = el_curtech;
	edifgbl.celltbl_sze = edifgbl.celltbl_cnt = 0;
	edifgbl.ports = edifgbl.free_ports = NOEDPORT;
	edifgbl.properties = edifgbl.free_properties = NOEDPROPERTY;
	edifgbl.firstnetport = edifgbl.lastnetport = edifgbl.free_netports = NOEDNETPORT;

	/* active database inits */
	edifgbl.current_cell = NONODEPROTO;
	edifgbl.current_node = NONODEINST;
	edifgbl.current_arc = NOARCINST;
	edifgbl.current_port = NOPORTPROTO;
	edifgbl.figure_group = NONODEPROTO;
	edifgbl.arc_type = NOARCPROTO;
	edifgbl.proto = NONODEPROTO;

	/* name inits */
	edifgbl.cell_reference[0] = '\0';
	edifgbl.port_reference[0] = '\0';
	edifgbl.instance_reference[0] = '\0';
	edifgbl.bundle_reference[0] = '\0';
	edifgbl.net_reference[0] = '\0';
	edifgbl.property_reference[0] = '\0';

	/* geometry inits */
	edifgbl.layer_ptr = 0;
	edifgbl.points = edifgbl.lastpt = NOEPT;
	edifgbl.pt_count = 0;
	edifgbl.cur_nametbl = NONAMETABLE;

	/* text inits */
	edifgbl.textheight = 0;
	edifgbl.justification = LOWERLEFT;
	edifgbl.visible = 1;
	edifgbl.save_text.points = NOEPT;
	edifgbl.save_text.pt_count = 0;
	edifgbl.save_text.string[0] = '\0';

	edifgbl.sheet_size = SHEET_DSIZE;
	edifgbl.sh_xpos = -1;
	edifgbl.sh_ypos = -1;

	kstack_ptr = 0;

	if (default_port == NOPORTPROTO) default_port = getportproto(sch_wirepinprim, "wire");
	default_iconport = sch_wirepinprim->firstportproto;
	if (default_busport == NOPORTPROTO) default_busport = getportproto(sch_buspinprim, "bus");
	if (default_input == NOPORTPROTO) default_input = getportproto(sch_offpageprim, "y");
	if (default_output == NOPORTPROTO) default_output = getportproto(sch_offpageprim, "a");

	if (EDIF_name_key == 0) EDIF_name_key = makekey("EDIF_name");
	if (EDIF_array_key == 0) EDIF_array_key = makekey("EDIF_array");
	if (EDIF_annotate_key == 0) EDIF_annotate_key = makekey("EDIF_annotate");

	/* now load the edif netlist */
	(void)initinfstr();
	(void)addstringtoinfstr("Reading ");
	(void)addstringtoinfstr(filename);
	msg = returninfstr();
	if (io_verbose < 0)
	{
		if (DiaInitDialog(&us_progressdialog) != 0)
		{
			xclose(edifgbl.edif_file);
			return(1);
		}
		DiaPercent(1, 0);
		DiaSetText(2, msg);
	} else ttyputmsg(msg);

	/* count the number of keywords */
	key_count = 0;
	while (edif_keywords[key_count].name != 0) key_count++;
	if (io_edload_edif(&edifgbl, key_count, edif_keywords))
	{
		ttyputerr("error: bad design, aborting netlist translation");
		if (io_verbose < 0) DiaDoneDialog();
		return(1);
	}
	if (edifgbl.errors || edifgbl.warnings)
		ttyputmsg("info: a total of %d errors, and %d warnings encountered during load",
			edifgbl.errors, edifgbl.warnings);

	xclose(edifgbl.edif_file);
	if (io_verbose < 0)
	{
		DiaPercent(1, 100);
		DiaSetText(2, "Cleaning Up...");
	}

	/* free the layer table */
	for (i = 0; i < edifgbl.layer_ptr; i++)
	{
		if (edifgbl.nametbl[i]->original != edifgbl.nametbl[i]->replace)
			efree(edifgbl.nametbl[i]->replace);
		efree(edifgbl.nametbl[i]->original);
		efree((char *)edifgbl.nametbl[i]);
	}
	/* note the master list is static */

	/* now the cell name table */
	for(i = 0; i < edifgbl.celltbl_cnt; i++)
	{
		if (edifgbl.celltbl[i]->original != edifgbl.celltbl[i]->replace)
			efree(edifgbl.celltbl[i]->replace);
		efree(edifgbl.celltbl[i]->original);
		efree((char *)edifgbl.celltbl[i]);
	}

	/* and the master list ... */
	if (edifgbl.celltbl_cnt != 0)
	{
		efree((char *)edifgbl.celltbl);
	} else
	{
		ttyputerr("error: no data");
		if (io_verbose < 0) DiaDoneDialog();
		return(1);
	}

	if (edifgbl.library->curnodeproto == NONODEPROTO)
		edifgbl.library->curnodeproto = edifgbl.library->firstnodeproto;

	for (np = edifgbl.library->firstnodeproto; np != NONODEPROTO; np = np->nextnodeproto)
		net_examinenodeproto(np);

	/* recompute bounds */
	(*el_curconstraint->solve)(NONODEPROTO);

	if (io_verbose < 0) DiaDoneDialog();
	return(0);
}

/*
 * Module: io_edload_edif
 * Function:  Will load the edif netlist into memory
 * Method:  Does a simple keyword lookup to load the lisp structured
 *		 EDIF language into Electric's database
 */
INTBIG io_edload_edif(EDIFGBL_PTR edifgbl, INTBIG key_count, EDIFKEY_PTR keywords)
{
	KSTATE *state;
	char *msg;
	char *token, word[WORD+1];
	INTBIG low, high, mid, change;
	INTBIG keycount;
	INTBIG savestack;

	/* setup the error return location */
	if ((msg = (char *)setjmp(edifgbl->env)) != NULL)
	{
		ttyputerr("line #%d: %s", edifgbl->lineno, msg);
		return 1;
	}

	savestack = -1;

	/* now read and parse the edif netlist */
	keycount = 0;
	while ((token = io_edget_keyword(edifgbl, word)) != NULL)
	{
		keycount++;
		if ((keycount%100) == 0)
		{
			if (io_verbose < 0) DiaPercent(1, filepos*100/filelen); else
				ttyputmsg("%d keywords read (%d%% done)", keycount, filepos*100/filelen);
		}

		/* locate the keyword, and execute the function */
		low = 0;
		high = key_count - 1;
		while (low <= high)
		{
			mid = (low + high) / 2;
			change = namesame(token, keywords[mid].name);
			if (change < 0) high = mid - 1; else
				if (change > 0) low = mid + 1; else
			{
				/* found the function, check state */
				if (keywords[mid].state != KNULL && edifgbl->state != KUNKNOWN )
				{
					state = keywords[mid].state;
					while (*state && edifgbl->state != *state) state++;
					if (edifgbl->state != *state)
					{
						ttyputerr("error, line #%d: illegal state for keyword <%s>",
							edifgbl->lineno, token);
						edifgbl->errors++;
					}
				}

				/* call the function */
				kstack[kstack_ptr++] = edifgbl->state;
				edifgbl->state = keywords[mid].next;
				if (savestack >= kstack_ptr)
				{
					savestack = -1;
					edifgbl->ignoreblock = 0;
				}
				if (!edifgbl->ignoreblock) (keywords[mid].func)(edifgbl);
				if (edifgbl->ignoreblock)
				{
					if (savestack == -1) savestack = kstack_ptr;
				}
				break;
			}
		}
		if (low > high)
		{
			ttyputerr("warning, line #%d: unknown keyword <%s>", edifgbl->lineno, token);
			edifgbl->warnings++;
			kstack[kstack_ptr++] = edifgbl->state;
			edifgbl->state = KUNKNOWN;
		}
	}
	if (edifgbl->state != KINIT)
	{
		ttyputerr("line #%d: unexpected end-of-file encountered", edifgbl->lineno);
		edifgbl->errors++;
	}
	return 0;
}

/* parser routines */
INTBIG io_ednoop(EDIFGBL_PTR edifgbl)
{
	return 0;
}

INTBIG io_edtechnology(EDIFGBL_PTR edifgbl)
{
	return 0;
}

INTBIG io_edfabricate(EDIFGBL_PTR edifgbl)
{
	INTBIG pos;
	char name[WORD+1];

	pos = edifgbl->layer_ptr;

	edifgbl->nametbl[pos] = (NAMETABLE_PTR)emalloc(sizeof(NAMETABLE), io_aid->cluster);
	if (edifgbl->nametbl[pos] == NONAMETABLE) RET_NOMEMORY();

	/* first get the original and replacement layers */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if ((edifgbl->nametbl[pos]->original = io_edstralloc(name)) == NULL) RET_NOMEMORY();
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if ((edifgbl->nametbl[pos]->replace = io_edstralloc(name)) == NULL) RET_NOMEMORY();
	edifgbl->nametbl[pos]->textheight = 0;
	edifgbl->nametbl[pos]->justification = LOWERLEFT;
	edifgbl->nametbl[pos]->visible = 1;

	/* now bump the position */
	edifgbl->layer_ptr++;
	return 0;
}

INTBIG io_edpage(EDIFGBL_PTR edifgbl)
{
	char view[WORD+1], cname[WORD+1];
	NODEPROTO *proto;

	/* check for the name */
	(void)io_edcheck_name(edifgbl);

	(void)sprintf(view, "p%d", ++edifgbl->pageno);

	/* locate this in the list of cells */
	for (proto = edifgbl->library->firstnodeproto; proto != NONODEPROTO; proto = proto->nextnodeproto)
	{
		if (!namesame(proto->cell->cellname, edifgbl->cell_name) &&
			!namesame(proto->cellview->sviewname, view)) break;
	}
	if (proto == NONODEPROTO)
	{
		/* allocate the cell */
		sprintf(cname, "%s{%s}", edifgbl->cell_name, view);
		if ((proto = newnodeproto(cname, edifgbl->library)) == NONODEPROTO)
			longjmp(edifgbl->env, (int)"no memory");
		proto->temp1 = 0;

		/* add the facet center node */
#ifdef ED_FACETCENTER
		if (newnodeinst(gen_facetcenterprim, gen_facetcenterprim->lowx, gen_facetcenterprim->highx,
			gen_facetcenterprim->lowy, gen_facetcenterprim->highy, 0, 0, proto) == NONODEINST)
		{
			ttyputerr("Could not create facet center in %s", cname);
		}
#endif
	}
	else if (proto->temp1) edifgbl->ignoreblock = 1;

	edifgbl->current_cell = proto;
	return 0;
}

INTBIG io_edfigure(EDIFGBL_PTR edifgbl)
{
	char layer[WORD+1];
	INTBIG low, high, change, mid;
	INTBIG pos;

	/* get the layer name */
	/* check for figuregroup override */
	if (*io_edpos_token(edifgbl) == LEFTP) return 0;
	if ((io_edget_token(edifgbl, layer, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	/* now look for this layer in the list of layers */
	low = 0;
	high = edifgbl->layer_ptr - 1;
	while (low <= high)
	{
		mid = (low + high) / 2;
		change = namesame(layer, edifgbl->nametbl[mid]->original);
		if (change < 0) high = mid - 1; else
			if (change > 0) low = mid + 1; else
		{
			/* found the layer */
			edifgbl->figure_group = edifgbl->nametbl[mid]->node;
			edifgbl->arc_type = edifgbl->nametbl[mid]->arc;
			edifgbl->textheight = edifgbl->nametbl[mid]->textheight;
			edifgbl->justification = edifgbl->nametbl[mid]->justification;
			edifgbl->visible = edifgbl->nametbl[mid]->visible;

			/* allow new definitions */
			edifgbl->cur_nametbl = edifgbl->nametbl[mid];
			break;
		}
	}
	if (low > high)
	{
		pos = edifgbl->layer_ptr;
		edifgbl->nametbl[pos] = (NAMETABLE_PTR)emalloc(sizeof(NAMETABLE), io_aid->cluster);
		if (edifgbl->nametbl[pos] == NONAMETABLE ||
			(edifgbl->nametbl[pos]->original = io_edstralloc(layer)) == NULL)
				longjmp(edifgbl->env, (int)"no memory");
		edifgbl->nametbl[pos]->replace = edifgbl->nametbl[pos]->original;
		edifgbl->figure_group = edifgbl->nametbl[pos]->node = art_boxprim;
		edifgbl->arc_type = edifgbl->nametbl[pos]->arc = sch_wirearc;
		edifgbl->textheight = edifgbl->nametbl[pos]->textheight = 0;
		edifgbl->justification = edifgbl->nametbl[pos]->justification = LOWERLEFT;
		edifgbl->visible = edifgbl->nametbl[pos]->visible = 1;

		/* allow new definitions */
		edifgbl->cur_nametbl = edifgbl->nametbl[pos];

		/* now sort the list */
		edifgbl->layer_ptr++;
		qsort(edifgbl->nametbl, edifgbl->layer_ptr, sizeof(NAMETABLE_PTR),
			(__cmp_func)io_edcompare_name);
	}
	return 0;
}

INTBIG io_edfigureGroupOverride(EDIFGBL_PTR edifgbl)
{
	return io_edgetlayer(edifgbl);
}

INTBIG io_edscale(EDIFGBL_PTR edifgbl)
{
	/* get the scale */
	edifgbl->val1 = io_edgetnumber(edifgbl);
	edifgbl->val2 = io_edgetnumber(edifgbl);
	return 0;
}

INTBIG io_edendType(EDIFGBL_PTR edifgbl)
{
	char type[WORD+1];

	/* get the endtype */
	if ((io_edget_token(edifgbl, type, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (!namesame(type, "EXTEND")) edifgbl->extend_end = TRUE;
	return 0;
}

INTBIG io_edcornerType(EDIFGBL_PTR edifgbl)
{
	char type[WORD+1];

	/* get the endtype */
	if ((io_edget_token(edifgbl, type, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (!namesame(type, "EXTEND")) edifgbl->extend_corner = TRUE;
	return 0;
}

INTBIG io_edstringDisplay(EDIFGBL_PTR edifgbl)
{
	INTBIG kptr;

	/* init the point lists */
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	edifgbl->visible = 1;
	edifgbl->justification = LOWERLEFT;
	edifgbl->textheight = 0;

	/* get the string, remove the quote */
	io_edget_delim(edifgbl, QUOTE);
	if ((io_edget_token(edifgbl, edifgbl->string, QUOTE)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	/* check for RENAME */
	if (kstack[kstack_ptr-1] != KRENAME) return 0;
	(void)strcpy(edifgbl->original, edifgbl->string);
	if (kstack[kstack_ptr - 2] == KARRAY) kptr = kstack_ptr - 3; else
		kptr = kstack_ptr - 2;
	switch (kstack[kptr])
	{
		case KCELL:
			(void)strcpy(edifgbl->cell_name, edifgbl->original);
			break;
		case KPORT:
			(void)strcpy(edifgbl->port_name, edifgbl->original);
			break;
		case KINSTANCE:
			(void)strcpy(edifgbl->instance_name, edifgbl->original);
			break;
		case KNET:
			(void)strcpy(edifgbl->net_name, edifgbl->original);
			break;
		case KNETBUNDLE:
			(void)strcpy(edifgbl->bundle_name, edifgbl->original);
			break;
		case KPROPERTY:
			(void)strcpy(edifgbl->property_name, edifgbl->original);
			break;
	}
	return 0;
}

INTBIG io_edtrue(EDIFGBL_PTR edifgbl)
{
	/* check previous keyword */
	if (kstack_ptr > 1 && kstack[kstack_ptr-1] == KVISIBLE)
	{
		edifgbl->visible = 1;
		if (edifgbl->cur_nametbl != NONAMETABLE)
			edifgbl->cur_nametbl->visible = 1;
	}
	return 0;
}

INTBIG io_edfalse(EDIFGBL_PTR edifgbl)
{
	/* check previous keyword */
	if (kstack_ptr > 1 && kstack[kstack_ptr-1] == KVISIBLE)
	{
		edifgbl->visible = 0;
		if (edifgbl->cur_nametbl != NONAMETABLE)
			edifgbl->cur_nametbl->visible = 0;
	}
	return 0;
}

INTBIG io_edgetlayer(EDIFGBL_PTR edifgbl)
{
	char layer[WORD+1];
	INTBIG low, high, change, mid;
	INTBIG pos;

	/* get the layer name */
	if ((io_edget_token(edifgbl, layer, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	/* now look for this layer in the list of layers */
	low = 0;
	high = edifgbl->layer_ptr - 1;
	while (low <= high)
	{
		mid = (low + high) / 2;
		change = namesame(layer, edifgbl->nametbl[mid]->original);
		if (change < 0) high = mid - 1; else
			if (change > 0) low = mid + 1; else
		{
			/* found the layer */
			edifgbl->figure_group = edifgbl->nametbl[mid]->node;
			edifgbl->arc_type = edifgbl->nametbl[mid]->arc;
			edifgbl->textheight = edifgbl->nametbl[mid]->textheight;
			edifgbl->justification = edifgbl->nametbl[mid]->justification;
			edifgbl->visible = edifgbl->nametbl[mid]->visible;
			break;
		}
	}
	if (low > high)
	{
#ifdef LAYERWARNING
		ttyputerr("warning, line #%d: unknown layer <%s>", edifgbl->lineno, layer);
		edifgbl->warnings++;
#endif

		/* insert and resort the list */
		pos = edifgbl->layer_ptr;

		edifgbl->nametbl[pos] = (NAMETABLE_PTR)emalloc(sizeof(NAMETABLE), io_aid->cluster);
		if (edifgbl->nametbl[pos] == NONAMETABLE ||
			(edifgbl->nametbl[pos]->original = io_edstralloc(layer)) == NULL)
				longjmp(edifgbl->env, (int)"no memory");

		edifgbl->nametbl[pos]->replace = edifgbl->nametbl[pos]->original;
		edifgbl->figure_group = edifgbl->nametbl[pos]->node = art_boxprim;
		edifgbl->arc_type = edifgbl->nametbl[pos]->arc = sch_wirearc;
		edifgbl->textheight = edifgbl->nametbl[pos]->textheight = 0;
		edifgbl->justification = edifgbl->nametbl[pos]->justification = LOWERLEFT;
		edifgbl->visible = edifgbl->nametbl[pos]->visible = 1;

		/* now sort the list */
		edifgbl->layer_ptr++;
		qsort(edifgbl->nametbl, edifgbl->layer_ptr, sizeof(NAMETABLE_PTR),
			(__cmp_func)io_edcompare_name);
	}
	return 0;
}

/* geometry routines */
INTBIG io_edpointList(EDIFGBL_PTR edifgbl)
{
	return 0;
}

INTBIG io_edpoint(EDIFGBL_PTR edifgbl)
{
	return 0;
}

INTBIG io_edjustify(EDIFGBL_PTR edifgbl)
{
	char val[WORD+1];

	/* get the textheight value of the point */
	if ((io_edget_token(edifgbl, val, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if (namesame(val, "UPPERLEFT"))
		edifgbl->justification = UPPERLEFT;
	else if (namesame(val, "UPPERCENTER"))
		edifgbl->justification = UPPERCENTER;
	else if (namesame(val, "UPPERRIGHT"))
		edifgbl->justification = UPPERRIGHT;
	else if (namesame(val, "CENTERLEFT"))
		edifgbl->justification = CENTERLEFT;
	else if (namesame(val, "CENTERCENTER"))
		edifgbl->justification = CENTERCENTER;
	else if (namesame(val, "CENTERRIGHT"))
		edifgbl->justification = CENTERRIGHT;
	else if (namesame(val, "LOWERLEFT"))
		edifgbl->justification = LOWERLEFT;
	else if (namesame(val, "LOWERCENTER"))
		edifgbl->justification = LOWERCENTER;
	else if (namesame(val, "LOWERRIGHT"))
		edifgbl->justification = LOWERRIGHT;
	else
	{
		ttyputerr("warning, line #%d: unknown keyword <%s>", edifgbl->lineno, val);
		edifgbl->warnings++;
		return 0;
	}

	if (edifgbl->cur_nametbl != NONAMETABLE)
		edifgbl->cur_nametbl->justification = edifgbl->justification;
	return 0;
}

INTBIG io_edtextHeight(EDIFGBL_PTR edifgbl)
{
	char val[WORD+1];

	/* get the textheight value of the point */
	if ((io_edget_token(edifgbl, val, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	edifgbl->textheight = (INTBIG)(atol(val) * edifgbl->scale);

	if (edifgbl->cur_nametbl != NONAMETABLE)
		edifgbl->cur_nametbl->textheight = edifgbl->textheight;

	return 0;
}

INTBIG io_edpt(EDIFGBL_PTR edifgbl)
{
	char xstr[WORD+1], ystr[WORD+1];
	EPT_PTR point;
	INTBIG x, y, s;

	/* get the x and y values of the point */
	if ((io_edget_token(edifgbl, xstr, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if ((io_edget_token(edifgbl, ystr, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (kstack_ptr > 1 && kstack[kstack_ptr-1] == KDELTA)
	{
		x = (INTBIG)(atol(xstr) * edifgbl->scale);
		y = (INTBIG)(atol(ystr) * edifgbl->scale);

		/* transform the points per orientation */
		switch (edifgbl->orientation)
		{
			case OR0:
				break;
			case OR90:
				s = x;
				x = -y;
				y = s;
				break;
			case OR180:
				x = -x;
				y = -y;
				break;
			case OR270:
				s = x;
				x = y;
				y = -s;
				break;
			case OMY:
				x = -x;
				break;
			case OMX:
				y = -y;
				break;
			case OMYR90:
				s = y;
				y = -x;
				x = -s;
				break;
			case OMXR90:
				s = x;
				x = y;
				y = s;
				break;
		}
		/* set the array deltas */
		if (edifgbl->deltapts == 0)
		{
			edifgbl->deltaxX = x;
			edifgbl->deltaxY = y;
		} else
		{
			edifgbl->deltayX = x;
			edifgbl->deltayY = y;
		}
		edifgbl->deltapts++;
	} else
	{
		/* allocate a point to read */
		if ((point = (EPT_PTR) emalloc(sizeof (EPT), io_aid->cluster)) == NOEPT) RET_NOMEMORY();

		/* and set the values */
		point->x = atol(xstr);
		if (point->x > 0) point->x = (INTBIG)(point->x * edifgbl->scale + 0.5); else
			point->x = (INTBIG)(point->x * edifgbl->scale - 0.5);
		point->y = atol(ystr);
		if (point->y > 0) point->y = (INTBIG)(point->y * edifgbl->scale + 0.5); else
			point->y = (INTBIG)(point->y * edifgbl->scale - 0.5);
		point->nextpt = NOEPT;

		/* add it to the list of points */
		if (edifgbl->points) edifgbl->lastpt->nextpt = point; else
			edifgbl->points = point;
		edifgbl->lastpt = point;
		edifgbl->pt_count++;
	}
	return 0;
}

INTBIG io_edorientation(EDIFGBL_PTR edifgbl)
{
	char orient[WORD+1];

	/* get the orientation keyword */
	if ((io_edget_token(edifgbl, orient, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (!namesame(orient, "R0")) edifgbl->orientation = OR0;
	else if (!namesame(orient, "R90")) edifgbl->orientation = OR90;
	else if (!namesame(orient, "R180")) edifgbl->orientation = OR180;
	else if (!namesame(orient, "R270")) edifgbl->orientation = OR270;
	else if (!namesame(orient, "MY")) edifgbl->orientation = OMY;
	else if (!namesame(orient, "MX")) edifgbl->orientation = OMX;
	else if (!namesame(orient, "MYR90")) edifgbl->orientation = OMYR90;
	else if (!namesame(orient, "MXR90")) edifgbl->orientation = OMXR90;
	else
	{
		ttyputerr("warning, line #%d: unknown orientation value <%s>", edifgbl->lineno, orient);
		edifgbl->warnings++;
	}
	return 0;
}

INTBIG io_edpathWidth(EDIFGBL_PTR edifgbl)
{
	char width[WORD+1];

	/* get the width string */
	if ((io_edget_token(edifgbl, width, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	edifgbl->path_width = atol(width);

	return 0;
}

INTBIG io_edrectangle(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	return 0;
}

INTBIG io_edcircle(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	return 0;
}

INTBIG io_edpath(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	edifgbl->path_width = 0;
	return 0;
}

INTBIG io_edpolygon(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	return 0;
}

INTBIG io_edopenShape(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	return 0;
}

INTBIG io_edfreeptlist(EDIFGBL_PTR edifgbl)
{
	while (edifgbl->points != NOEPT)
	{
		edifgbl->lastpt = edifgbl->points->nextpt;
		efree((char *) edifgbl->points);
		edifgbl->points = edifgbl->lastpt;
		edifgbl->pt_count --;
	}
	if (edifgbl->pt_count != 0)
		longjmp(edifgbl->env, (int)"internal error, point list mismatch");
	return 0;
}

INTBIG io_edfreesavedptlist(EDIFGBL_PTR edifgbl)
{
	EPT_PTR lastpt;
	while (edifgbl->save_text.points != NOEPT)
	{
		lastpt = edifgbl->save_text.points->nextpt;
		efree((char *)edifgbl->save_text.points);
		edifgbl->save_text.points = lastpt;
		edifgbl->save_text.pt_count --;
	}
	if (edifgbl->save_text.pt_count != 0)
		longjmp(edifgbl->env, (int)"internal error, point list mismatch");
	return 0;
}

INTBIG io_edprogram(EDIFGBL_PTR edifgbl)
{
	char program[WORD+1];

	if ((io_edget_token(edifgbl, program, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if (!namesamen(&program[1], "VIEWlogic", 9))
	{
		edifgbl->vendor = EVVIEWLOGIC;
	} else if (!namesamen(&program[1], "edifout", 7))
	{
		edifgbl->vendor = EVCADENCE;
	}
	return 0;
}

INTBIG io_edunit(EDIFGBL_PTR edifgbl)
{
	char type[WORD+1];

	if ((io_edget_token(edifgbl, type, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if (kstack[kstack_ptr-1] == KSCALE && !namesame(type, "DISTANCE"))
	{
		/* get the scale */
		edifgbl->dbunit = edifgbl->val1;
		edifgbl->meters = edifgbl->val2;
		edifgbl->scale = edifgbl->meters/edifgbl->dbunit;
#if 0
		/* recompute the scale */
		/* if schematic technology, 4,000 db = 0.1 inch or 0.00254 meters */
		edifgbl->scale = edifgbl->scale * (2.0 * (double)edifgbl->technology->deflambda) / 0.00254;
#else
		edifgbl->scale = edifgbl->scale * 2.0 *
			scaletodispunit(edifgbl->technology->deflambda, DISPUNITINCH) * 10.0 *	/* 1/10 inch */
				scalefromdispunit(100.0, DISPUNITCM);								/* units per meter */
#endif
	}
	return 0;
}

INTBIG io_edlibrary(EDIFGBL_PTR edifgbl)
{
	char *name, nbuf[WORD+1];

	/* get the name of the library */
	name = nbuf;
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
#ifdef EDIFLIBNAME
	/* retrieve the library */
	if ((edifgbl->library = getlibrary(name)) == NOLIBRARY)
	{
		/* create the library */
		edifgbl->library = newlibrary(name, name);
		if (edifgbl->library == NOLIBRARY)
			longjmp(edifgbl->env, (int)"could not create library");
		ttyputmsg("Reading library %s", name);
	} else
	{
		/* flag existing cells */
		for (np = edifgbl->library->firstnodeproto; np != NONODEPROTO; np = np->nextnodeproto)
			np->temp1 = 1;
		ttyputmsg("Updating library %s", name);
	}
#endif
	return 0;
}

/* cell definition routines */
INTBIG io_eddesign(EDIFGBL_PTR edifgbl)
{
	char name[WORD+1];

	/* get the name of the cell */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	return 0;
}

/* note: cells are considered primitive unless they contain instance
   information, i.e. (contents  (instance ... , this cell will become
   a COMPLEX cell.
 */
INTBIG io_edcell(EDIFGBL_PTR edifgbl)
{
	edifgbl->active_view = VNULL;
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->pageno = 0;
	edifgbl->sh_xpos = edifgbl->sh_ypos = -1;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->cell_reference, edifgbl->name);
		(void)strcpy(edifgbl->cell_name, edifgbl->name);
	}
	return 0;
}

/* view: Indicates a new view of the cell
 * 	 just clears the active view flag
 */
INTBIG io_edview(EDIFGBL_PTR edifgbl)
{
	edifgbl->active_view = VNULL;
	return 0;
}

/* viewType:  Indicates the view style for this cell, ie
 *	BEHAVIOR, DOCUMENT, GRAPHIC, LOGICMODEL, MASKLAYOUT, NETLIST,
 *  	PCBLAYOUT, SCHEMATIC, STRANGER, SYMBOLIC
 * we are only concerned about the viewType NETLIST.
 */
INTBIG io_edviewType(EDIFGBL_PTR edifgbl)
{
	char name[WORD+1], cname[WORD+1], view[WORD+1];
	NODEPROTO *proto;

	/* get the viewType */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (!namesame(name, "BEHAVIOR")) edifgbl->active_view = VBEHAVIOR;
	else if (!namesame(name, "DOCUMENT")) edifgbl->active_view = VDOCUMENT;
	else if (!namesame(name, "GRAPHIC"))
	{
		edifgbl->active_view = VGRAPHIC;
		strcpy(view, "");
	}
	else if (!namesame(name, "LOGICMODEL")) edifgbl->active_view = VLOGICMODEL;
	else if (!namesame(name, "MASKLAYOUT"))
	{
		edifgbl->active_view = VMASKLAYOUT;
		strcpy(view, "lay");
	}
	else if (!namesame(name, "NETLIST"))
	{
		edifgbl->active_view = VNETLIST;
		strcpy(view, "sch");
	}
	else if (!namesame(name, "PCBLAYOUT")) edifgbl->active_view = VPCBLAYOUT;
	else if (!namesame(name, "SCHEMATIC"))
	{
		edifgbl->active_view = VSCHEMATIC;
		strcpy(view, "sch");
	}
	else if (!namesame(name, "STRANGER")) edifgbl->active_view = VSTRANGER;
	else if (!namesame(name, "SYMBOLIC")) edifgbl->active_view = VSYMBOLIC;

	/* immediately allocate MASKLAYOUT and VGRAPHIC viewtypes */
	if (edifgbl->active_view == VMASKLAYOUT || edifgbl->active_view == VGRAPHIC ||
		edifgbl->active_view == VNETLIST || edifgbl->active_view == VSCHEMATIC)
	{
		/* locate this in the list of cells */
		for (proto = edifgbl->library->firstnodeproto; proto != NONODEPROTO;
			proto = proto->nextnodeproto)
		{
			if (!namesame(proto->cell->cellname, edifgbl->cell_name) &&
				!namesame(proto->cellview->sviewname, view)) break;
		}
		if (proto == NONODEPROTO)
		{
			/* allocate the cell */
			sprintf(cname, "%s{%s}", edifgbl->cell_name, view);
			if ((proto = newnodeproto(cname, edifgbl->library)) == NONODEPROTO)
				longjmp(edifgbl->env, (int)"no memory");
			proto->temp1 = 0;
#ifdef ED_FACETCENTER
			/* add the facet center node */
			if (newnodeinst(gen_facetcenterprim, gen_facetcenterprim->lowx,
				gen_facetcenterprim->highx, gen_facetcenterprim->lowy, gen_facetcenterprim->highy,
					0, 0, proto) == NONODEINST)
			{
				ttyputerr("Could not create facet center in %s", cname);
			}
#endif
		}
		else if (proto->temp1) edifgbl->ignoreblock = 1;

		edifgbl->current_cell = proto;
	}
	else edifgbl->current_cell = NONODEPROTO;

	/* add the name to the celltbl */
	if (edifgbl->celltbl_cnt == edifgbl->celltbl_sze)
	{
		edifgbl->celltbl_sze += 100;
		if (edifgbl->celltbl_cnt == 0)
		{
			edifgbl->celltbl = (NAMETABLE_PTR *)emalloc(sizeof (NAMETABLE_PTR) *
				edifgbl->celltbl_sze, io_aid->cluster);
			if (edifgbl->celltbl == (NAMETABLE_PTR *)0) RET_NOMEMORY();
		} else
		{
			NAMETABLE_PTR *newtbl;
			long i;

			newtbl = (NAMETABLE_PTR *)emalloc(sizeof (NAMETABLE_PTR) * edifgbl->celltbl_sze,
				io_aid->cluster);
			if (newtbl == (NAMETABLE_PTR *)0) RET_NOMEMORY();
			for(i=0; i<edifgbl->celltbl_cnt; i++) newtbl[i] = edifgbl->celltbl[i];
			edifgbl->celltbl = newtbl;
		}
	}
	edifgbl->celltbl[edifgbl->celltbl_cnt] = (NAMETABLE_PTR)emalloc((unsigned) sizeof (NAMETABLE),
		io_aid->cluster);
	if (edifgbl->celltbl[edifgbl->celltbl_cnt] == NONAMETABLE) RET_NOMEMORY();
	edifgbl->celltbl[edifgbl->celltbl_cnt]->original = io_edstralloc(edifgbl->cell_reference);
	if (edifgbl->celltbl[edifgbl->celltbl_cnt]->original == NULL) RET_NOMEMORY();
	edifgbl->celltbl[edifgbl->celltbl_cnt]->replace = io_edstralloc(edifgbl->cell_name);
	if (edifgbl->celltbl[edifgbl->celltbl_cnt]->replace == NULL) RET_NOMEMORY();

	/* now sort by name */
	qsort(edifgbl->celltbl, ++(edifgbl->celltbl_cnt), sizeof (NAMETABLE_PTR),
		(__cmp_func)io_edcompare_name);
	return 0;
}

/* contents:  Indicates additional information about higher level blocks
 * initially verifies the appropriate state, ie
 * 	(view XXX (viewType Netlist)
 *	  (contents ...
 */
INTBIG io_edcontents(EDIFGBL_PTR edifgbl)
{
	return 0;
}

/* port: Define connection point for the cell, must be in INTERFACE */
INTBIG io_edport(EDIFGBL_PTR edifgbl)
{
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->direction = INPUT;
	edifgbl->port_reference[0] = '\0';
	edifgbl->isarray = 0;
	edifgbl->arrayx = edifgbl->arrayy = 1;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->port_reference, edifgbl->name);
		(void)strcpy(edifgbl->port_name, edifgbl->name);
	}
	return 0;
}

/* direction:  Set the direction of a port */
INTBIG io_eddirection(EDIFGBL_PTR edifgbl)
{
	char name[WORD+1];

	/* get the direction */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	if (!namesame(name, "INOUT")) edifgbl->direction = INOUT; else
		if (!namesame(name, "INPUT")) edifgbl->direction = INPUT; else
			if (!namesame(name, "OUTPUT")) edifgbl->direction = OUTPUT;
	return 0;
}

/* instance definition routines */
/* instance:  Indicates a new instance within the cell
 */
INTBIG io_edinstance(EDIFGBL_PTR edifgbl)
{
	/* set the current geometry type */
	io_edfreeptlist(edifgbl);
	edifgbl->proto = NONODEPROTO;
	edifgbl->geometry = GINSTANCE;
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	edifgbl->isarray = 0;
	edifgbl->arrayx = edifgbl->arrayy = 1;
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->instance_reference[0] = '\0';
	edifgbl->current_node = NONODEINST;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->instance_reference, edifgbl->name);
		(void)strcpy(edifgbl->instance_name, edifgbl->name);
	}
	return 0;
}

INTBIG io_edarray(EDIFGBL_PTR edifgbl)
{
	/* note array is a special process integer value function */
	edifgbl->isarray = 1;
	edifgbl->arrayx = edifgbl->arrayy = 0;
	edifgbl->deltaxX = edifgbl->deltaxY = 0;
	edifgbl->deltayX = edifgbl->deltayY = 0;
	edifgbl->deltapts = 0;
	if (io_edcheck_name(edifgbl))
	{
		switch (kstack[kstack_ptr-1])
		{
			case KCELL:
				(void)strcpy(edifgbl->cell_reference, edifgbl->name);
				(void)strcpy(edifgbl->cell_name, edifgbl->name);
				break;
			case KPORT:
				(void)strcpy(edifgbl->port_name, edifgbl->name);
				(void)strcpy(edifgbl->port_reference, edifgbl->name);
				break;
			case KINSTANCE:
				(void)strcpy(edifgbl->instance_name, edifgbl->name);
				(void)strcpy(edifgbl->instance_reference, edifgbl->name);
				break;
			case KNET:
				(void)strcpy(edifgbl->net_reference, edifgbl->name);
				(void)strcpy(edifgbl->net_name, edifgbl->name);
				break;
			case KPROPERTY:
				(void)strcpy(edifgbl->property_reference, edifgbl->name);
				(void)strcpy(edifgbl->property_name, edifgbl->name);
				break;
		}
	}
	return 0;
}

INTBIG io_eddelta(EDIFGBL_PTR edifgbl)
{
	edifgbl->deltaxX = edifgbl->deltaxY = 0;
	edifgbl->deltayX = edifgbl->deltayY = 0;
	edifgbl->deltapts = 0;
	return 0;
}

/* cellRef:  determines the cell type of an instance
 */
INTBIG io_edcellRef(EDIFGBL_PTR edifgbl)
{
	char name[WORD+1], view[WORD+1];
	NODEPROTO *proto;
	INTBIG low, high, change, mid;

	/* get the name of the cell */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	switch (edifgbl->active_view)
	{
		case VMASKLAYOUT:
			strcpy(view, "lay");
			break;
		default:
			if (kstack[kstack_ptr - 1] == KDESIGN) strcpy(view, "p1"); else
				strcpy(view, "ic");
	}
	if (edifgbl->vendor == EVVIEWLOGIC && !namesame(name, "SPLITTER"))
	{
		edifgbl->proto = NONODEPROTO;
		return 0;
	}

	/* look for this cell name in the cell list */
	low = 0;
	high = edifgbl->celltbl_cnt - 1;
	while (low <= high)
	{
		mid = (low + high) / 2;
		change = namesame(name, edifgbl->celltbl[mid]->original);
		if (change < 0) high = mid - 1; else
			if (change > 0) low = mid + 1; else
		{
			/* found the name */
			(void)strcpy(name, edifgbl->celltbl[mid]->replace);
			break;
		}
	}
	if (low > high)
	{
		ttyputerr("could not find cellRef <%s>", name);
	}

	/* now look for this cell in the list of cells, if not found create it */
	/* locate this in the list of cells */
	for (proto = edifgbl->library->firstnodeproto; proto != NONODEPROTO; proto = proto->nextnodeproto)
	{
		if (!namesame(proto->cell->cellname, name) &&
			!namesame(proto->cellview->sviewname, view)) break;
	}
	if (proto == NONODEPROTO)
	{
		/* allocate the cell */
		if (view[0] != '\0') sprintf(&name[strlen(name)], "{%s}", view);
		if ((proto = newnodeproto(name, edifgbl->library)) == NONODEPROTO)
			longjmp(edifgbl->env, (int)"no memory");
		proto->temp1 = 0;
#ifdef ED_FACETCENTER
		/* add the facet center node */
		if (newnodeinst(gen_facetcenterprim, gen_facetcenterprim->lowx,
			gen_facetcenterprim->highx, gen_facetcenterprim->lowy, gen_facetcenterprim->highy,
				0, 0, proto) == NONODEINST)
		{
			ttyputerr("Could not create facet center in %s", name);
		}
#endif
	}
	/* set the parent */
	edifgbl->proto = proto;
	return 0;
}

INTBIG io_edinteger(EDIFGBL_PTR edifgbl)
{
	char value[WORD+1];

	if ((io_edget_token(edifgbl, value, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");

	edifgbl->pval.integer = atoi(value);
	edifgbl->ptype = PINTEGER;
	return 0;
}

INTBIG io_ednumber(EDIFGBL_PTR edifgbl)
{
	edifgbl->pval.number = (float)io_edgetnumber(edifgbl);
	edifgbl->ptype = PNUMBER;
	return 0;
}

INTBIG io_edstring(EDIFGBL_PTR edifgbl)
{
	char *token, value[WORD+1];

	if (*(token = io_edpos_token(edifgbl)) != LEFTP && *token != RIGHTP)
	{
		if ((io_edget_token(edifgbl, value, 0)) == NULL)
			longjmp(edifgbl->env, (int)"unexpected end-of-file");

		value[strlen(value) - 1] = '\0';
		if ((edifgbl->pval.string = (char *)emalloc(strlen(value), io_aid->cluster)) == NULL)
			longjmp(edifgbl->env, (int)"no memory");

		(void)strcpy(edifgbl->pval.string, &value[1]);
		edifgbl->ptype = PSTRING;
	}
	return 0;
}

INTBIG io_edproperty(EDIFGBL_PTR edifgbl)
{
	edifgbl->property_reference[0] = '\0';
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->ptype = PUNKNOWN;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->property_reference, edifgbl->name);
		(void)strcpy(edifgbl->property_name, edifgbl->name);
	}
	return 0;
}

/* net:  Define a net name followed by edifgbl and local pin list
 *		(net NAME
 *			(joined
 *			(portRef NAME (instanceRef NAME))
 *			(portRef NAME)))
 */
INTBIG io_ednet(EDIFGBL_PTR edifgbl)
{
	edifgbl->net_reference[0] = '\0';
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->current_arc = NOARCINST;
	edifgbl->current_node = NONODEINST;
	edifgbl->current_port = NOPORTPROTO;
	edifgbl->firstnetport = edifgbl->lastnetport = NOEDNETPORT;
	edifgbl->isarray = 0;
	edifgbl->arrayx = edifgbl->arrayy = 1;
	if (kstack[kstack_ptr-2] != KNETBUNDLE) edifgbl->geometry = GNET;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->net_reference, edifgbl->name);
		(void)strcpy(edifgbl->net_name, edifgbl->name);
	}
	return 0;
}

INTBIG io_ednetBundle(EDIFGBL_PTR edifgbl)
{
	edifgbl->geometry = GBUS;
	edifgbl->bundle_reference[0] = '\0';
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	edifgbl->isarray = 0;
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->bundle_reference, edifgbl->name);
		(void)strcpy(edifgbl->bundle_name, edifgbl->name);
	}
	return 0;
}

INTBIG io_edjoined(EDIFGBL_PTR edifgbl)
{
	return 0;
}

/* portRef:  specifies a pin on a net */
INTBIG io_edportRef(EDIFGBL_PTR edifgbl)
{
	edifgbl->port_reference[0] = '\0';
	edifgbl->instance_reference[0] = '\0';
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->port_reference, edifgbl->name);
	}

	/* allocate a netport */
	if (io_edallocnetport(edifgbl))
		longjmp(edifgbl->env, (int)"no memory");
	return 0;
}

/* instanceRef:  identifies the name of the instance attached to the net */
INTBIG io_edinstanceRef(EDIFGBL_PTR edifgbl)
{
	edifgbl->instance_reference[0] = '\0';
	if (io_edcheck_name(edifgbl))
	{
		(void)strcpy(edifgbl->instance_reference, edifgbl->name);
	}
	return 0;
}

/* array member function, determines a specific element of an array */
INTBIG io_edmember(EDIFGBL_PTR edifgbl)
{
	edifgbl->memberx = -1; /* no member */
	edifgbl->membery = -1; /* no member */
	if (io_edcheck_name(edifgbl))
	{
		switch (kstack[kstack_ptr-1])
		{
			case KPORTREF:
				(void)strcpy(edifgbl->port_reference, edifgbl->name);
				break;
			case KINSTANCEREF:
				(void)strcpy(edifgbl->instance_reference, edifgbl->name);
				break;
		}
	}
	return 0;
}

INTBIG io_edname(EDIFGBL_PTR edifgbl)
{
	INTBIG kptr;

	if (kstack[kstack_ptr-1] == KARRAY || kstack[kstack_ptr-1] == KMEMBER) kptr = kstack_ptr-2; else
		kptr = kstack_ptr - 1;
	if (io_edcheck_name(edifgbl))
	{
		switch (kstack[kptr])
		{
			case KCELL:
				(void)strcpy(edifgbl->cell_reference, edifgbl->name);
				(void)strcpy(edifgbl->cell_name, edifgbl->name);
				break;
			case KPORTIMPLEMENTATION:		/* added by smr */
			case KPORT:
				(void)strcpy(edifgbl->port_name, edifgbl->name);
			case KPORTREF:
				(void)strcpy(edifgbl->port_reference, edifgbl->name);
				break;
			case KINSTANCE:
				(void)strcpy(edifgbl->instance_name, edifgbl->name);
			case KINSTANCEREF:
				(void)strcpy(edifgbl->instance_reference, edifgbl->name);
				break;
			case KNET:
				(void)strcpy(edifgbl->net_reference, edifgbl->name);
				(void)strcpy(edifgbl->net_name, edifgbl->name);
				break;
			case KPROPERTY:
				(void)strcpy(edifgbl->property_reference, edifgbl->name);
				(void)strcpy(edifgbl->property_name, edifgbl->name);
				break;
		}

		/* init the point lists */
		io_edfreeptlist(edifgbl);
		edifgbl->points = edifgbl->lastpt = NOEPT;
		edifgbl->orientation = OR0;
		edifgbl->visible = 1;
		edifgbl->justification = LOWERLEFT;
		edifgbl->textheight = 0;
		(void)strcpy(edifgbl->string, edifgbl->name);
	}
	return 0;
}

INTBIG io_edrename(EDIFGBL_PTR edifgbl)
{
	char name[WORD+1];
	INTBIG kptr;

	/* get the name of the object */
	if ((io_edget_token(edifgbl, name, 0)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	(void)strcpy(edifgbl->name, (name[0] == '&') ? &name[1] : name);

	/* and the original name */
	if (*io_edpos_token(edifgbl) == LEFTP)
	{
		/* must be stringDisplay, copy name to original */
		(void)strcpy(edifgbl->original, edifgbl->name);
	} else
	{
		if  (io_edget_token(edifgbl, name, 0) == NULL)
			longjmp(edifgbl->env, (int)"unexpected end-of-file");

		/* copy name without quotes */
		(void)strcpy(edifgbl->original, &name[1]);
		edifgbl->original[strlen(name)-2] = '\0';
	}
	if (kstack[kstack_ptr - 1]  == KNAME) kptr = kstack_ptr - 1; else
		kptr = kstack_ptr;
	if (kstack[kptr - 1] == KARRAY) kptr = kptr - 2; else
		kptr = kptr -1;
	switch (kstack[kptr])
	{
		case KCELL:
			(void)strcpy(edifgbl->cell_reference, edifgbl->name);
			(void)strcpy(edifgbl->cell_name, edifgbl->original);
			break;
		case KPORT:
			(void)strcpy(edifgbl->port_reference, edifgbl->name);
			(void)strcpy(edifgbl->port_name, edifgbl->original);
			break;
		case KINSTANCE:
			(void)strcpy(edifgbl->instance_reference, edifgbl->name);
			(void)strcpy(edifgbl->instance_name, edifgbl->original);
			break;
		case KNETBUNDLE:
			(void)strcpy(edifgbl->bundle_reference, edifgbl->name);
			(void)strcpy(edifgbl->bundle_name, edifgbl->original);
			break;
		case KNET:
			(void)strcpy(edifgbl->net_reference, edifgbl->name);
			(void)strcpy(edifgbl->net_name, edifgbl->original);
			break;
		case KPROPERTY:
			(void)strcpy(edifgbl->property_reference, edifgbl->name);
			(void)strcpy(edifgbl->property_name, edifgbl->original);
			break;
	}
	return 0;
}

/* symbol routines and schematic routines */
/* boundingBox - describes the bounding box of a schematic symbol */
INTBIG io_edboundingBox(EDIFGBL_PTR edifgbl)
{
	edifgbl->figure_group = art_openeddottedpolygonprim;
	return 0;
}

/* symbol - describes a schematic symbol */
INTBIG io_edsymbol(EDIFGBL_PTR edifgbl)
{
	char cname[WORD+1];
	NODEPROTO *proto;
	INTBIG center[2];

	edifgbl->active_view = VSYMBOL;

	/* locate this in the list of cells */
	for (proto = edifgbl->library->firstnodeproto; proto != NONODEPROTO; proto = proto->nextnodeproto)
	{
		if (!namesame(proto->cell->cellname, edifgbl->cell_name) &&
			!namesame(proto->cellview->sviewname, "ic")) break;
	}
	if (proto == NONODEPROTO)
	{
		/* allocate the cell */
		sprintf(cname, "%s{ic}", edifgbl->cell_name);
		if ((proto = newnodeproto(cname, edifgbl->library)) == NONODEPROTO)
			longjmp(edifgbl->env, (int)"no memory");
		proto->userbits |= WANTNEXPAND;
		proto->temp1 = 0;
#ifdef ED_FACETCENTER
		/* add the facet center node */
		hw = edifgbl->technology->deflambda/2;
		if (newnodeinst(gen_facetcenterprim, -hw, hw, -hw, hw, 0, 0, proto) == NONODEINST)
			ttyputerr("Could not create facet center in %s", cname);
#endif
		center[0] = center[1] = 0;
		(void)setvalkey((INTBIG)proto, VNODEPROTO, el_prototype_center,
			(INTBIG)center, VINTEGER|VISARRAY|(2<<VLENGTHSH));

	}
	else if (proto->temp1) edifgbl->ignoreblock = 1;

	edifgbl->current_cell = proto;
	edifgbl->figure_group = NONODEPROTO;
	return 0;
}

INTBIG io_edportImplementation(EDIFGBL_PTR edifgbl)
{
	/* set the current geometry type */
	io_edfreeptlist(edifgbl);
	edifgbl->proto = sch_wirepinprim;
	edifgbl->geometry = GPIN;
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	edifgbl->isarray = 0;
	edifgbl->arrayx = edifgbl->arrayy = 1;
	edifgbl->name[0] = edifgbl->original[0] = '\0';
	(void)io_edcheck_name(edifgbl);
	return 0;
}

INTBIG io_edconnectLocation(EDIFGBL_PTR edifgbl)
{
	return 0;
}

INTBIG io_edinterface(EDIFGBL_PTR edifgbl)
{
	NODEPROTO *np;
	char nodename[WORD+1];

	/* create schematic page 1 to represent all I/O for this schematic */
	/* locate this in the list of cells */
	for (np = edifgbl->library->firstnodeproto; np != NONODEPROTO; np = np->nextnodeproto)
	{
		if (!namesame(np->cell->cellname, edifgbl->cell_name) &&
			!namesame(np->cellview->sviewname, "sch")) break;
	}
	if (np == NONODEPROTO)
	{
		/* allocate the cell */
		sprintf(nodename, "%s{sch}", edifgbl->cell_name);
		if ((edifgbl->current_cell = newnodeproto(nodename, edifgbl->library)) == NONODEPROTO)
			longjmp(edifgbl->env, (int)"no memory");
		edifgbl->current_cell->temp1 = 0;
	}
	else edifgbl->current_cell = np;

	/* now set the current position in the schematic page */
	edifgbl->ipos = edifgbl->bpos = edifgbl->opos = 0;
	return 0;
}

INTBIG io_eddot(EDIFGBL_PTR edifgbl)
{
	io_edfreeptlist(edifgbl);
	edifgbl->points = edifgbl->lastpt = NOEPT;
	edifgbl->orientation = OR0;
	return 0;
}

double io_edgetnumber(EDIFGBL_PTR edifgbl)
{
	char value[WORD+1];
	INTBIG matissa, exponent;

	if (io_edget_token(edifgbl, value, 0) == NULL)
	{
		longjmp(edifgbl->env, (int)"no integer value");
	}
	if (value[0] == LEFTP)
	{
		/* must be in e notation */
		if (io_edget_token(edifgbl, value, 0) == NULL ||
			namesame(value, "e"))
			longjmp(edifgbl->env, (int)"illegal number value");

		/* now the matissa */
		if (io_edget_token(edifgbl, value, 0) == NULL)
			longjmp(edifgbl->env, (int)"no matissa value");
		matissa = atoi(value);

		/* now the exponent */
		if (io_edget_token(edifgbl, value, 0) == NULL)
			longjmp(edifgbl->env, (int)"no exponent value");
		exponent = atoi(value);
		io_edget_delim(edifgbl, RIGHTP);
		return (double)matissa * pow(10.0, (double)exponent);
	}
	return (double) atoi(value);
}

INTBIG io_edcheck_name(EDIFGBL_PTR edifgbl)
{
	char *pp;
	char name[WORD+1];

	if (*(pp = io_edpos_token(edifgbl)) != LEFTP && *pp != RIGHTP)
	{
		if (io_edget_token(edifgbl, name, 0) == NULL)
			longjmp(edifgbl->env, (int)"Illegal name");
		(void)strcpy(edifgbl->name, (name[0] == '&') ? &name[1] : name);
		return 1;
	}
	return 0;
}

static void io_edmake_electric_name(char *ref, char *name, int x, int y)
{
}

/* general utilities ... */
static void io_edcheck_busnames(ARCINST *ai, char *base)
{
	PORTARCINST *pai;
	NODEINST *ni;
	INTSML i, l;
	VARIABLE *var;
	char *name, newname[WORD+1];

	if (ai->proto != sch_busarc)
	{
		/* verify the name */
		if ((var = getvalkey((INTBIG)ai, VARCINST, VSTRING, el_arc_name)) != NOVARIABLE &&
			((name = (char *)var->addr) != NULL))
		{
			l = strlen(base);
			if (!namesamen(name, base, l) && isdigit(name[l]))
			{
				(void)sprintf(newname, "%s[%s]", base, &name[l]);
				setvalkey((INTBIG)ai, VARCINST, el_arc_name, (INTBIG)newname, VSTRING);
			}
		}
	}
	ai->temp1 = 1;
	for (i = 0; i < 2; i++)
	{
		ni = ai->end[i].nodeinst;
		if (((ni->proto->userbits & NFUNCTION) >> NFUNCTIONSH) == NPPIN)
		{
			/* scan through this nodes portarcinst's */
			for (pai = ni->firstportarcinst; pai != NOPORTARCINST; pai = pai->nextportarcinst)
			{
				if (pai->conarcinst->temp1 == 0)
					io_edcheck_busnames(pai->conarcinst, base);
			}
		}
	}
}

/* pop the keyword state stack, called when ')' is encountered. Note this function
   needs to broken into a simple function call structure. */
#define MAXBUSPINS 256
INTBIG io_edpop_stack(EDIFGBL_PTR edifgbl)
{
	INTBIG index, key;
	char nodename[WORD+1], portname[WORD+1], basename[WORD+1], *name, orig[WORD+1];
	NODEPROTO *np, *nnp;
	ARCPROTO *ap, *lap;
	ARCINST *ai;
	PORTPROTO *ppt, *fpp, *lpp, *pp, *lastport;
	NODEINST *ni, *lastpin, *fni, *lni;
	EPT_PTR point;
	EDPROPERTY_PTR property, nproperty;
	CELL *cell;
	INTBIG lx, ly, hx, hy, cx, cy, cnt, Ix, Iy, gx, gy, xoff, yoff, descript;
	INTBIG *trace, *pt, pts[26], fcnt, tcnt, psx, psy;
	INTBIG width, height, x[3], y[3], radius, trans;
	double ar, so;
	VARIABLE *var;
	INTBIG count, user_max, i, j, *layer_numbers, dup, fbus, tbus;
	XARRAY rot;
	char **layers;
	EDPORT_PTR eport, neport;
	EDNETPORT_PTR nport, nnport;
	/* bus connection variables */
	PORTPROTO *fpps[MAXBUSPINS], *tpps[MAXBUSPINS];
	NODEINST *fnis[MAXBUSPINS], *tnis[MAXBUSPINS];

	if (kstack_ptr)
	{
		if (!edifgbl->ignoreblock)
		{
			switch (edifgbl->state)
			{
			case KFIGURE:
				edifgbl->cur_nametbl = NONAMETABLE;
				edifgbl->visible = 1;
				edifgbl->justification = LOWERLEFT;
				edifgbl->textheight = 0;
				break;
			case KBOUNDINGBOX:
				edifgbl->figure_group = NONODEPROTO;
				break;
			case KTECHNOLOGY:
				var = getval((INTBIG)edifgbl->technology, VTECHNOLOGY, VINTEGER|VISARRAY,
					"IO_gds_layer_numbers");
				if (var != NOVARIABLE)
				{
					layer_numbers = (INTBIG *)var->addr;
					count = getlength(var);
					var = getval((INTBIG)edifgbl->technology, VTECHNOLOGY, VSTRING|VISARRAY,
						"TECH_layer_names");
					if (var != NOVARIABLE)
					{
						layers = (char **) var->addr;
						user_max = edifgbl->layer_ptr;
						/* for all layers assign their GDS number */
						for (i=0; i<count; i++)
						{
							if (layer_numbers[i] != 0)
							{
								/* search for this layer */
								for (j=0; j<user_max; j++)
								{
									if (!namesame(edifgbl->nametbl[j]->replace, layers[i])) break;
								}
								if (user_max == j)
								{
									/* add to the list */
									if ((edifgbl->nametbl[edifgbl->layer_ptr] = (NAMETABLE_PTR)
										emalloc(sizeof(NAMETABLE), io_aid->cluster)) == NONAMETABLE) RET_NOMEMORY();
									if ((edifgbl->nametbl[edifgbl->layer_ptr]->replace = io_edstralloc(layers[i])) == NULL)
										RET_NOMEMORY();
									sprintf(orig, "layer_%d", layer_numbers[i]);
									if ((edifgbl->nametbl[edifgbl->layer_ptr]->original = io_edstralloc(orig)) == NULL)
										RET_NOMEMORY();
									edifgbl->nametbl[edifgbl->layer_ptr]->textheight = 0;
									edifgbl->nametbl[edifgbl->layer_ptr]->justification = LOWERLEFT;
									edifgbl->layer_ptr++;
								}
							}
						}
					}
				}

				/* sort the layer list */
				qsort(edifgbl->nametbl, edifgbl->layer_ptr, sizeof(NAMETABLE_PTR),
					(__cmp_func)io_edcompare_name);

				/* now look for nodes to map MASK layers to */
				for (index = 0; index < edifgbl->layer_ptr; index++)
				{
					sprintf(nodename, "%s-node", edifgbl->nametbl[index]->replace);
					for (np = edifgbl->technology->firstnodeproto; np != NONODEPROTO;
						np = np->nextnodeproto)
					{
						if (!namesame(nodename, np->primname)) break;
					}
					if (np == NONODEPROTO)
					{
						np = art_boxprim;
					}
					for (ap = edifgbl->technology->firstarcproto; ap != NOARCPROTO;
						ap = ap->nextarcproto)
					{
						if (!namesame(ap->protoname, edifgbl->nametbl[index]->replace))
							break;
					}
					edifgbl->nametbl[index]->node = np;
					edifgbl->nametbl[index]->arc = ap;
				}
				break;
			case KINTERFACE:
				if (edifgbl->active_view == VNETLIST)
				{
					/* create a black-box symbol at the current scale */
					np = edifgbl->current_cell;
					(void)sprintf(nodename, "%s{ic}", np->cell->cellname);
					if ((nnp = io_edmakeiconfacet(np->firstportproto, np->cell->cellname,
						nodename, edifgbl->technology, edifgbl->library)) == NONODEPROTO)
					{
						ttyputerr("error, line #%d: could not create icon <%s>",
							edifgbl->lineno, nodename);
						edifgbl->errors++;
					} else
					{
						/* now compute the bounds of this cell */
						db_boundfacet(nnp, &nnp->lowx, &nnp->highx, &nnp->lowy, &nnp->highy);
					}
				}
				break;
			case KVIEW:

				if (edifgbl->vendor == EVVIEWLOGIC && edifgbl->active_view != VNETLIST)
				{
					/* fixup incorrect bus nets */
					for (ai = edifgbl->current_cell->firstarcinst; ai != NOARCINST;
						ai = ai->nextarcinst)
					{
						ai->temp1 = ai->temp2 = 0;
					}

					/* now scan for BUS nets, and verify all wires connected to bus */
					for (ai = edifgbl->current_cell->firstarcinst; ai != NOARCINST;
						ai = ai->nextarcinst)
					{
						if (ai->temp1 == 0 && ai->proto == sch_busarc)
						{
							/* get name of arc */
							if ((var = getvalkey((INTBIG)ai, VARCINST, VSTRING, el_arc_name)) !=
								NOVARIABLE && ((name = (char *)var->addr) != NULL))
							{
								/* create the basename */
								(void)strcpy(basename, name);
								if ((name = strrchr(basename, '[')) != NULL) *name = '\0';

								/* expand this arc, and locate non-bracketed names */
								io_edcheck_busnames(ai, basename);
							}
						}
					}
				}

				/* move the port list to the free list */
				for (eport = edifgbl->ports; eport != NOEDPORT; eport = neport)
				{
					neport = eport->next;
					eport->next = edifgbl->free_ports;
					edifgbl->free_ports = eport;
					efree(eport->name);
					efree(eport->reference);
				}
				edifgbl->ports = NOEDPORT;
				break;
			case KPORT:
				if (io_edallocport(edifgbl)) return 1;
				eport = edifgbl->ports;
				np = edifgbl->current_cell;
				switch (eport->direction)
				{
					case INPUT:
						cx = 0;
						cy = edifgbl->ipos;
						edifgbl->ipos += INCH;
						fpp = default_input;
						break;
					case INOUT:
						cx = 3*INCH;
						cy = edifgbl->bpos;
						edifgbl->bpos += INCH;
						fpp = default_input;
						break;
					case OUTPUT:
						cx = 6*INCH;
						cy = edifgbl->opos;
						edifgbl->opos += INCH;
						fpp = default_output;
						break;
				}

				/* now create the off-page reference */
				defaultnodesize(sch_offpageprim, &psx, &psy);
				ni = newnodeinst(sch_offpageprim,
					cx + (sch_offpageprim->lowx+sch_offpageprim->highx-psx)/2,
					cx + (sch_offpageprim->lowx+sch_offpageprim->highx+psx)/2,
					cy + (sch_offpageprim->lowy+sch_offpageprim->highy-psy)/2,
					cy + (sch_offpageprim->lowy+sch_offpageprim->highy+psy)/2, 0,0, np);
				if (ni == NONODEINST)
				{
					ttyputerr("error, line #%d: could not create external port",
						edifgbl->lineno);
					edifgbl->errors++;
					break;
				}

				/* now create the port */
				if ((ppt = newportproto(ni->parent, ni, fpp, eport->name)) == NOPORTPROTO)
				{
					ttyputerr("error, line #%d: could not create port <%s>",
						edifgbl->lineno, eport->name);
					edifgbl->errors++;
				} else
				{
					switch (eport->direction)
					{
						case INPUT:
							ppt->userbits = (ppt->userbits & ~STATEBITS) | INPORT;
							break;
						case OUTPUT:
							ppt->userbits = (ppt->userbits & ~STATEBITS) | OUTPORT;
							break;
						case INOUT:
							ppt->userbits = (ppt->userbits & ~STATEBITS) | BIDIRPORT;
							break;
					}
				}
				edifgbl->port_reference[0] = '\0';

				/* move the property list to the free list */
				for (property = edifgbl->properties; property != NOEDPROPERTY; property = nproperty)
				{
					key = makekey(property->name);
					switch (property->type)
					{
						case PINTEGER:
							var = setvalkey((INTBIG)ppt, VPORTPROTO, key, (INTBIG)property->val.integer,
								VINTEGER);
							break;
						case PNUMBER:
							var = setvalkey((INTBIG)ppt, VPORTPROTO, key, (INTBIG)property->val.number,
								VFLOAT);
							break;
						case PSTRING:
							var = setvalkey((INTBIG)ppt, VPORTPROTO, key, (INTBIG)property->val.string,
								VSTRING);
							break;
					}
					nproperty = property->next;
					property->next = edifgbl->free_properties;
					edifgbl->free_properties = property;
					efree(property->name);
					if (property->type == PSTRING) efree(property->val.string);
				}
				edifgbl->properties = NOEDPROPERTY;
				break;
			case KINSTANCE:
				if (edifgbl->active_view == VNETLIST)
				{
					for (Ix = 0; Ix < edifgbl->arrayx; Ix++)
					{
						for (Iy = 0; Iy < edifgbl->arrayy; Iy++)
						{
							/* create this instance in the current sheet */
							width = edifgbl->proto->highx - edifgbl->proto->lowx;
							height = edifgbl->proto->highy - edifgbl->proto->lowy;
							width = (width + INCH - 1) / INCH;
							height = (height + INCH - 1) / INCH;

							/* verify room for the icon */
							if (edifgbl->sh_xpos != -1)
							{
								if ((edifgbl->sh_ypos + (height + 1))  >=
									EDIF_sheet_bounds[edifgbl->sheet_size].height)
								{
									edifgbl->sh_ypos = 1;
									if ((edifgbl->sh_xpos += edifgbl->sh_offset) >=
										EDIF_sheet_bounds[edifgbl->sheet_size].width)
											edifgbl->sh_xpos = edifgbl->sh_ypos = -1; else
												edifgbl->sh_offset = 2;
								}
							}
							if (edifgbl->sh_xpos == -1)
							{
								/* create the new page */
								(void)sprintf(nodename, "%s{p%d}", edifgbl->cell_name,
									++edifgbl->pageno);
								if ((edifgbl->current_cell = np = newnodeproto(nodename,
									edifgbl->library)) == NONODEPROTO)
								{
									longjmp(edifgbl->env, (int)"no memory");
								}
								np->temp1 = 0;
								edifgbl->sh_xpos = edifgbl->sh_ypos = 1;
								edifgbl->sh_offset = 2;
							}

							/* create this instance */
							/* find out where true center moves */
							cx = ((edifgbl->proto->lowx+edifgbl->proto->highx) >> 1) +
								((edifgbl->sh_xpos -
								(EDIF_sheet_bounds[edifgbl->sheet_size].width >> 1)) * INCH);
							cy = ((edifgbl->proto->lowy+edifgbl->proto->highy) >> 1) +
								((edifgbl->sh_ypos -
								(EDIF_sheet_bounds[edifgbl->sheet_size].height >> 1)) * INCH);
							edifgbl->current_node = ni = newnodeinst(edifgbl->proto,
								edifgbl->proto->lowx + cx,
								edifgbl->proto->highx + cx,
								edifgbl->proto->lowy + cy,
								edifgbl->proto->highy + cy,
								io_edgettrans(edifgbl->orientation),
								io_edgetrot(edifgbl->orientation),
								edifgbl->current_cell);
							if (ni == NONODEINST)
							{
								ttyputerr("error, line #%d: could not create instance",
									edifgbl->lineno);
								edifgbl->errors++;
								break;
							} else
							{
								if (edifgbl->proto->userbits & WANTNEXPAND)
								{
									ni->userbits |= NEXPAND;
								}

								/* update the current position */
								if ((width + 2) > edifgbl->sh_offset)
								{
									edifgbl->sh_offset = width + 2;
								}
								if ((edifgbl->sh_ypos += (height + 1) ) >=
									EDIF_sheet_bounds[edifgbl->sheet_size].height)
								{
									edifgbl->sh_ypos = 1;
									if ((edifgbl->sh_xpos += edifgbl->sh_offset) >=
										EDIF_sheet_bounds[edifgbl->sheet_size].width)
										edifgbl->sh_xpos = edifgbl->sh_ypos = -1; else
											edifgbl->sh_offset = 2;
								}

								/* name the instance */
								if (edifgbl->instance_reference[0] != '\0')
								{
									/* if single element or array with no offset */
									/* construct the representative extended EDIF name (includes [...]) */
									if ((edifgbl->arrayx == 1 && edifgbl->arrayy == 1) ||
										(edifgbl->deltaxX == 0 && edifgbl->deltaxY == 0 &&
										edifgbl->deltayX == 0 && edifgbl->deltayY == 0))
											(void)strcpy(nodename, edifgbl->instance_reference);
										/* if array in the x dimension */
									else if (edifgbl->arrayx > 1)
									{
										if (edifgbl->arrayy > 1)
											(void)sprintf(nodename, "%s[%d,%d]", edifgbl->instance_reference,
												Ix, Iy);
										else
											(void)sprintf(nodename, "%s[%d]", edifgbl->instance_reference, Ix);
									}
									/* if array in the y dimension */
									else if (edifgbl->arrayy > 1)
										(void)sprintf(nodename, "%s[%d]", edifgbl->instance_reference, Iy);

									/* check for array element descriptor */
									if (edifgbl->arrayx > 1 || edifgbl->arrayy > 1)
									{
										/* array descriptor is of the form index:index:range index:index:range */
										(void)sprintf(basename, "%d:%d:%d %d:%d:%d", Ix,
											(edifgbl->deltaxX == 0 && edifgbl->deltayX == 0) ?
											edifgbl->arrayx-1:Ix, edifgbl->arrayx, Iy,
											(edifgbl->deltaxY == 0 && edifgbl->deltayY == 0) ?
											edifgbl->arrayy-1:Iy, edifgbl->arrayy);
										var = setvalkey((INTBIG)ni, VNODEINST, EDIF_array_key, (INTBIG)basename,
											VSTRING);

									}

									/* now set the name of the component (note that Electric allows any string
									 * of characters as a name, this name is open to the user, for ECO and other
									 * consistancies, the EDIF_name is saved on a variable)
									 */
									if (!namesame(edifgbl->instance_reference, edifgbl->instance_name))
									{
										var = setvalkey((INTBIG)ni, VNODEINST, el_node_name,
											(INTBIG)nodename, VSTRING|VDISPLAY);
									} else
									{
										/* now add the original name as the displayed name (only to the first element) */
										if (Ix == 0 && Iy == 0)
											var = setvalkey((INTBIG)ni, VNODEINST, el_node_name,
												(INTBIG)edifgbl->instance_name, VSTRING|VDISPLAY);
										/* now save the EDIF name (not displayed) */
										var = setvalkey((INTBIG)ni, VNODEINST, EDIF_name_key,
											(INTBIG)nodename, VSTRING);
									}
								}
							}
						}
					}
				}

				/* move the property list to the free list */
				for (property = edifgbl->properties; property != NOEDPROPERTY; property = nproperty)
				{
					if (edifgbl->current_node != NONODEINST)
					{
						key = makekey(property->name);
						switch (property->type)
						{
							case PINTEGER:
								var = setvalkey((INTBIG)edifgbl->current_node, VNODEINST, key,
									(INTBIG)property->val.integer, VINTEGER);
								break;
							case PNUMBER:
								var = setvalkey((INTBIG)edifgbl->current_node, VNODEINST, key,
									(INTBIG)property->val.number, VFLOAT);
								break;
							case PSTRING:
								var = setvalkey((INTBIG)edifgbl->current_node, VNODEINST, key,
									(INTBIG)property->val.string, VSTRING);
								break;
						}
					}
					nproperty = property->next;
					property->next = edifgbl->free_properties;
					edifgbl->free_properties = property;
					efree(property->name);
					if (property->type == PSTRING) efree(property->val.string);
				}
				edifgbl->properties = NOEDPROPERTY;
				edifgbl->instance_reference[0] = '\0';
				edifgbl->current_node = NONODEINST;
				io_edfreesavedptlist(edifgbl);
				break;
			case KNET:
				/* move the property list to the free list */
				for (property = edifgbl->properties; property != NOEDPROPERTY; property = nproperty)
				{
					if (edifgbl->current_arc != NOARCINST)
					{
						key = makekey(property->name);
						switch (property->type)
						{
							case PINTEGER:
								var = setvalkey((INTBIG)edifgbl->current_arc, VARCINST, key,
									(INTBIG)property->val.integer, VINTEGER);
								break;
							case PNUMBER:
								var = setvalkey((INTBIG)edifgbl->current_arc, VARCINST, key,
									(INTBIG)property->val.number, VFLOAT);
								break;
							case PSTRING:
								var = setvalkey((INTBIG)edifgbl->current_arc, VARCINST, key,
									(INTBIG)property->val.string, VSTRING);
								break;
						}
					}
					nproperty = property->next;
					property->next = edifgbl->free_properties;
					edifgbl->free_properties = property;
					efree(property->name);
					if (property->type == PSTRING) efree(property->val.string);
				}
				edifgbl->properties = NOEDPROPERTY;
				edifgbl->net_reference[0] = '\0';
				edifgbl->current_arc = NOARCINST;
				if (edifgbl->geometry != GBUS)
					edifgbl->geometry = GUNKNOWN;
				io_edfreesavedptlist(edifgbl);
				break;
			case KNETBUNDLE:
				edifgbl->bundle_reference[0] = '\0';
				edifgbl->current_arc = NOARCINST;
				edifgbl->geometry = GUNKNOWN;
				io_edfreesavedptlist(edifgbl);
				break;
			case KPROPERTY:
				if (edifgbl->active_view == VNETLIST)
				{
					/* add as a variable to the current object */
					i = 0;
					switch (kstack[kstack_ptr - 1])
					{
						case KINTERFACE:
							/* add to the {sch} view nodeproto */
							for (np = edifgbl->library->firstnodeproto; np != NONODEPROTO;
								np = np->nextnodeproto)
							{
								if (!namesame(np->cell->cellname, edifgbl->cell_name) &&
									!namesame(np->cellview->sviewname, "sch")) break;
							}
							if (np == NONODEPROTO)
							{
								/* allocate the cell */
								sprintf(nodename, "%s{sch}", edifgbl->cell_name);
								if ((np = newnodeproto(nodename, edifgbl->library)) == NONODEPROTO)
									longjmp(edifgbl->env, (int)"no memory");
								np->temp1 = 0;
							}
							i = (INTBIG) np;
							j = VNODEPROTO;
							break;
						case KINSTANCE:
						case KNET:
						case KPORT:
							i = -1;
							break;
						default:
							i = 0;
							break;
					}
					if (i > 0)
					{
						key = makekey(edifgbl->property_reference);
						switch (edifgbl->ptype)
						{
							case PINTEGER:
								var = setvalkey(i, j, key, (INTBIG)edifgbl->pval.integer, VINTEGER);
								break;
							case PNUMBER:
								var = setvalkey(i, j, key, (INTBIG)edifgbl->pval.number, VFLOAT);
								break;
							case PSTRING:
								var = setvalkey(i, j, key, (INTBIG)edifgbl->pval.string, VSTRING);
								break;
						}
					} else if (i == -1)
					{
						/* add to the current property list, will be added latter */
						switch (edifgbl->ptype)
						{
							case PINTEGER:
								(void)io_edallocproperty(edifgbl, edifgbl->property_reference, edifgbl->ptype,
									(INTBIG)edifgbl->pval.integer, 0.0, NULL);
								break;
							case PNUMBER:
								(void)io_edallocproperty(edifgbl, edifgbl->property_reference, edifgbl->ptype,
									0, edifgbl->pval.number, NULL);
								break;
							case PSTRING:
								(void)io_edallocproperty(edifgbl, edifgbl->property_reference, edifgbl->ptype,
									0, 0.0, edifgbl->pval.string);
								break;
						}
					}
				}
				edifgbl->property_reference[0] = '\0';
				io_edfreesavedptlist(edifgbl);
				break;
			case KPORTIMPLEMENTATION:
				edifgbl->geometry = GUNKNOWN;
				break;
			case KTRANSFORM:
				if (kstack_ptr <= 1 || kstack[kstack_ptr-1] != KINSTANCE)
				{
					io_edfreeptlist(edifgbl);
					break;
				}

				/* get the corner offset */
				if (edifgbl->pt_count == 1 && edifgbl->proto != NONODEPROTO)
				{
					for (Ix = 0; Ix < edifgbl->arrayx; Ix++)
					{
						lx = edifgbl->points->x + Ix * edifgbl->deltaxX;
						ly = edifgbl->points->y + Ix * edifgbl->deltaxY;
						for (Iy = 0; Iy < edifgbl->arrayy; Iy++)
						{
							/* create the node instance */
							/* make rotations about the origin not center */
							/* build transformation for this operation */
							makeangle(io_edgetrot(edifgbl->orientation),
								io_edgettrans(edifgbl->orientation), rot);

							/* find out where true center moves */
							cx = (edifgbl->proto->lowx+edifgbl->proto->highx)/2;
							cy = (edifgbl->proto->lowy+edifgbl->proto->highy)/2;
							xform(cx, cy, &gx, &gy, rot);

							/* now calculate the delta movement of the center */
							cx = gx - cx;
							cy = gy - cy;

							edifgbl->current_node = ni = newnodeinst(edifgbl->proto,
								lx + edifgbl->proto->lowx + cx,
								lx + edifgbl->proto->highx + cx,
								ly + edifgbl->proto->lowy + cy,
								ly + edifgbl->proto->highy + cy,
								io_edgettrans(edifgbl->orientation),
								io_edgetrot(edifgbl->orientation),
								edifgbl->current_cell);
							if (ni == NONODEINST)
							{
								ttyputerr("error, line #%d: could not create instance",
									edifgbl->lineno);
								edifgbl->errors++;

								/* and exit for loop */
								Ix = edifgbl->arrayx;
								Iy = edifgbl->arrayy;
							} else
							{
								if (edifgbl->proto->userbits & WANTNEXPAND)
								{
									ni->userbits |= NEXPAND;
								}
							}
							if (edifgbl->geometry == GPIN && lx == 0 && ly == 0)
							{
								/* determine an appropriate port name */
								(void)strcpy(portname, edifgbl->port_name);

								/* check if port already exists (will occur for multiple portImplementation statements */
								for (dup = 0, (void)strcpy(basename, portname);
									(ppt = getportproto(ni->parent, portname)) != NOPORTPROTO;
									dup++)
								{
									if (dup == 0) pp = ppt;
									(void)sprintf(portname, "%s_%d", basename, dup+1);
								}

								/* only once */
								Ix = edifgbl->arrayx;
								Iy = edifgbl->arrayy;
								if ((ppt = newportproto(ni->parent, ni, default_iconport,
									portname)) == NOPORTPROTO)
								{
									ttyputerr("error, line #%d: could not create port <%s>",
										edifgbl->lineno, portname);
									edifgbl->errors++;
								} else
								{
									/* locate the direction of the port */
									for (eport = edifgbl->ports; eport != NOEDPORT; eport = eport->next)
									{
										if (!namesame(eport->reference, edifgbl->port_reference))
										{
											/* set the direction */
											switch (eport->direction)
											{
												case INPUT:
													ppt->userbits = (ppt->userbits & ~STATEBITS) | INPORT;
													break;
												case OUTPUT:
													ppt->userbits = (ppt->userbits & ~STATEBITS) | OUTPORT;
													break;
												case INOUT:
													ppt->userbits = (ppt->userbits & ~STATEBITS) | BIDIRPORT;
													break;
											}
											break;
										}
									}
								}
							} else
							{
								/* name the instance */
								if (edifgbl->instance_reference[0] != '\0')
								{
									/* if single element or array with no offset */
									/* construct the representative extended EDIF name (includes [...]) */
									if ((edifgbl->arrayx == 1 && edifgbl->arrayy == 1) ||
										(edifgbl->deltaxX == 0 && edifgbl->deltaxY == 0 &&
										edifgbl->deltayX == 0 && edifgbl->deltayY == 0))
										(void)strcpy(nodename, edifgbl->instance_reference);
										/* if array in the x dimension */
									else if (edifgbl->arrayx > 1)
									{
										if (edifgbl->arrayy > 1)
											(void)sprintf(nodename, "%s[%d,%d]", edifgbl->instance_reference,
												Ix, Iy);
										else
											(void)sprintf(nodename, "%s[%d]", edifgbl->instance_reference, Ix);
									}
									/* if array in the y dimension */
									else if (edifgbl->arrayy > 1)
										(void)sprintf(nodename, "%s[%d]", edifgbl->instance_reference, Iy);

									/* check for array element descriptor */
									if (edifgbl->arrayx > 1 || edifgbl->arrayy > 1)
									{
										/* array descriptor is of the form index:index:range index:index:range */
										(void)sprintf(basename, "%d:%d:%d %d:%d:%d", Ix,
											(edifgbl->deltaxX == 0 && edifgbl->deltayX == 0) ? edifgbl->arrayx-1:Ix,
											edifgbl->arrayx, Iy,
											(edifgbl->deltaxY == 0 && edifgbl->deltayY == 0) ? edifgbl->arrayy-1:Iy,
											edifgbl->arrayy);
										var = setvalkey((INTBIG)ni, VNODEINST, EDIF_array_key, (INTBIG)basename,
											VSTRING);

									}

									/* now set the name of the component (note that Electric allows any string
			   						 * of characters as a name, this name is open to the user, for ECO and other
									 * consistancies, the EDIF_name is saved on a variable)
									 */
									if (!namesame(edifgbl->instance_reference, edifgbl->instance_name))
									{
										var = setvalkey((INTBIG)ni, VNODEINST, el_node_name,
											(INTBIG)nodename, VSTRING|VDISPLAY);
									} else
									{
										/* now add the original name as the displayed name (only to the first element) */
										if (Ix == 0 && Iy == 0)
											var = setvalkey((INTBIG)ni, VNODEINST, el_node_name,
												(INTBIG)edifgbl->instance_name, VSTRING|VDISPLAY);
										/* now save the EDIF name (not displayed) */
										var = setvalkey((INTBIG)ni, VNODEINST, EDIF_name_key,
											(INTBIG)nodename, VSTRING);
									}

									/* now check for saved name attributes */
									if (edifgbl->save_text.points != NOEPT)
									{
										/* now set the position, relative to the center of the current object */
										xoff = edifgbl->save_text.points->x - ((ni->highx+ni->lowx)>>1);
										yoff = edifgbl->save_text.points->y - ((ni->highy+ni->lowy)>>1);

										/* convert to quarter lambda units */
										xoff = 4*xoff / edifgbl->technology->deflambda;
										yoff = 4*yoff / edifgbl->technology->deflambda;

										/*
										 * determine the size of text, 0.0278 in == 2 points or 36 (2xpixels) == 1 in
										 * fonts range from TXT4P to TXT20P, in 2 pixel increments
										 */
										if (edifgbl->save_text.textheight == 0) i = TXTSMALL; else
										{
											i = (int)((double)edifgbl->save_text.textheight * 3.6 /
												(double)edifgbl->technology->deflambda + 0.5) - 2;
											if (i < TXT4P) i = TXT4P;
											if (i > TXT20P) i = TXT20P;
										}

										descript = var->textdescript;
										descript &= ~(VTPOSITION|VTXOFF|VTYOFF|VTSIZE);
										descript |= (abs(xoff) << VTXOFFSH) | (xoff < 0 ? VTXOFFNEG : 0) |
											(abs(yoff) << VTYOFFSH) | (yoff < 0 ? VTYOFFNEG : 0) |
											(i << VTSIZESH);
										switch (edifgbl->save_text.justification)
										{
											case UPPERLEFT:
												descript |=  VTPOSUPRIGHT;
												break;
											case UPPERCENTER:
												descript |=  VTPOSUP;
												break;
											case UPPERRIGHT:
												descript |=  VTPOSUPLEFT;
												break;
											case CENTERLEFT:
												descript |=  VTPOSRIGHT;
												break;
											case CENTERCENTER:
												descript |=  VTPOSCENT;
												break;
											case CENTERRIGHT:
												descript |=  VTPOSLEFT;
												break;
											case LOWERLEFT:
												descript |=  VTPOSDOWNRIGHT;
												break;
											case LOWERCENTER:
												descript |=  VTPOSDOWN;
												break;
											case LOWERRIGHT:
												descript |=  VTPOSDOWNLEFT;
												break;
										}
										var->textdescript = descript;
									}
								}
							}
							if (edifgbl->deltayX == 0 && edifgbl->deltayY == 0) break;

							/* bump the y delta and x deltas */
							lx += edifgbl->deltayX;
							ly += edifgbl->deltayY;
						}
						if (edifgbl->deltaxX == 0 && edifgbl->deltaxY == 0) break;
					}
				}
				io_edfreeptlist(edifgbl);
				break;
			case KPORTREF:
				/* check for the last pin */
				fni = edifgbl->current_node;
				fpp = edifgbl->current_port;
				if (edifgbl->port_reference[0] != '\0')
				{
					/* For internal pins of an instance, determine the base port location and
					 * other pin assignments
					 */
					if (edifgbl->instance_reference[0] != '\0')
					{
						(void)strcpy(nodename, edifgbl->instance_reference);

						/* locate the node and and port */
						if (edifgbl->active_view == VNETLIST)
						{
							/* scan all pages for this nodeinst */
							for (cell = edifgbl->library->firstcell; cell != NOCELL; cell = cell->nextcell)
								if (!namesame(cell->cellname, edifgbl->cell_name)) break;
							if (cell == NOCELL)
							{
								ttyputmsg("error, line #%d: could not locate cell (%s)",
									edifgbl->lineno, edifgbl->cell_name);
								break;
							}

							/* scan for this instance in all sheets */
							for (np = cell->firstincell; np != NONODEPROTO; np = np->nextincell)
							{
								for (ni = np->firstnodeinst; ni != NONODEINST; ni = ni->nextnodeinst)
								{
									if ((var = getvalkey((INTBIG)ni, VNODEINST, VSTRING, EDIF_name_key)) == NOVARIABLE &&
										(var = getvalkey((INTBIG)ni, VNODEINST, VSTRING, el_node_name)) == NOVARIABLE)
											continue;
									if (!namesame((char *)var->addr, nodename)) break;
								}
								if (ni != NONODEINST) break;
							}
							ap = gen_unroutedarc;
						} else
						{
							/* net always references the current page */
							for (ni = edifgbl->current_cell->firstnodeinst; ni != NONODEINST;
								ni = ni->nextnodeinst)
							{
								if ((var = getvalkey((INTBIG)ni, VNODEINST, VSTRING, EDIF_name_key)) == NOVARIABLE &&
									(var = getvalkey((INTBIG)ni, VNODEINST, VSTRING, el_node_name)) == NOVARIABLE)
										continue;
								if (!namesame((char *)var->addr, nodename)) break;
							}
							if (edifgbl->isarray == 0) ap = sch_wirearc;
							else ap = sch_busarc;
						}
						if (ni == NONODEINST)
						{
							(void)ttyputmsg("error, line #%d: could not locate node (%s)",
								edifgbl->lineno, nodename);
							break;
						}
						/* locate the port for this portref */
						for (pp = ni->proto->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
						{
							if (!namesame(pp->protoname, edifgbl->port_reference)) break;
						}
						if (pp == NOPORTPROTO)
						{
							(void)ttyputmsg("error, line #%d: could not locate port (%s) on node (%s)",
								edifgbl->lineno, edifgbl->port_reference, nodename);
							break;
						}

						/* we have both, set global variable */
						edifgbl->current_node = ni;
						edifgbl->current_port = pp;
						np = edifgbl->current_cell;

						/* create extensions for net labels on single pin nets (externals), and
						 * placeholder for auto-routing later
						 */
						if (edifgbl->active_view == VNETLIST)
						{
							/* route all pins with an extension */
							if (ni != NONODEINST)
							{
								portposition(ni, pp, &hx, &hy);
								lx = hx;
								ly = hy;
								switch (pp->userbits&STATEBITS)
								{
									case INPORT:
										lx = hx - (INCH/10);
										break;
									case BIDIRPORT:
										ly = hy - (INCH/10);
										break;
									case OUTPORT:
										lx = hx + (INCH/10);
										break;
								}

								/* need to create a destination for the wire */
								if (edifgbl->isarray != 0)
								{
									if ((lni = newnodeinst(sch_buspinprim, lx + sch_buspinprim->lowx,
										lx + sch_buspinprim->highx, ly + sch_buspinprim->lowy,
										ly + sch_buspinprim->highy, 0, 0, np)) == NONODEINST)
									{
										ttyputmsg("error, line#%d: could not create bus pin", edifgbl->lineno);
										break;
									}
									lpp = default_busport;
									lap = sch_busarc;
								} else
								{
									if ((lni = newnodeinst(sch_wirepinprim, lx + sch_wirepinprim->lowx,
										lx + sch_wirepinprim->highx, ly + sch_wirepinprim->lowy,
										ly + sch_wirepinprim->highy, 0, 0, np)) == NONODEINST)
									{
										ttyputmsg("error, line#%d: could not create wire pin", edifgbl->lineno);
										break;
									}
									lpp = default_port;
									lap = sch_wirearc;
								}
								if ((edifgbl->current_arc = newarcinst(lap, defaultarcwidth(lap), CANTSLIDE,
									lni, lpp, lx, ly, ni, pp, hx, hy, np)) == NOARCINST)
										ttyputmsg("error, line #%d: could not create auto-path",
											edifgbl->lineno);
								else
									io_ednamearc(edifgbl, edifgbl->current_arc);
							}
						}
					} else
					{
						/* external port reference, look for a off-page reference in {sch} with this
			 			  port name */
						for (cell = edifgbl->library->firstcell; cell != NOCELL; cell = cell->nextcell)
							if (!namesame(cell->cellname, edifgbl->cell_name)) break;
						if (cell == NOCELL)
						{
							ttyputmsg("error, line #%d: could not locate cell (%s)",
								edifgbl->lineno, edifgbl->cell_name);
							break;
						}

						/* scan for this page in the cell */
						for (np = cell->firstincell; np != NONODEPROTO; np = np->nextincell)
						{
							if (!namesame(np->cellview->sviewname, "sch")) break;
						}
						if (np == NONODEPROTO)
						{
							ttyputmsg("error, line #%d: could not locate top level schematic",
								edifgbl->lineno);
							break;
						}

						/* now look for an instance with the correct port name */
						for (pp = np->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
						{
							if (!namesame(pp->protoname, edifgbl->port_reference)) break;
						}
						if (pp == NOPORTPROTO)
						{
							ttyputmsg("error, line #%d: could not locate port", edifgbl->lineno);
							break;
						}
						fni = pp->subnodeinst;
						fpp = pp->subportproto;
						portposition(fni, fpp, &lx, &ly);

						/* determine x position by original placement */
						switch (pp->userbits&STATEBITS)
						{
							case INPORT:
								lx = 1*INCH;
								break;
							case BIDIRPORT:
								lx = 4*INCH;
								break;
							case OUTPORT:
								lx = 5*INCH;
								break;
						}

						/* need to create a destination for the wire */
						if (edifgbl->isarray != 0)
						{
							if ((ni = newnodeinst(sch_buspinprim, lx + sch_buspinprim->lowx,
								lx + sch_buspinprim->highx, ly + sch_buspinprim->lowy,
								ly + sch_buspinprim->highy, 0, 0, np)) == NONODEINST)
							{
								ttyputmsg("error, line#%d: could not create bus pin",
									edifgbl->lineno);
								break;
							}
							pp = default_busport;
						} else
						{
							if ((ni = newnodeinst(sch_wirepinprim, lx + sch_wirepinprim->lowx,
								lx + sch_wirepinprim->highx, ly + sch_wirepinprim->lowy,
								ly + sch_wirepinprim->highy, 0, 0, np)) == NONODEINST)
							{
								ttyputmsg("error, line#%d: could not create wire pin",
									edifgbl->lineno);
								break;
							}
							pp = default_port;
						}
						if (edifgbl->isarray == 0) ap = sch_wirearc; else
							ap = sch_busarc;
					}

					/* now connect if we have from node and port */
					if (fni != NONODEINST && fpp != NOPORTPROTO)
					{
						/* locate the position of the new ports */
						portposition(fni, fpp, &lx, &ly);
						portposition(ni, pp, &hx, &hy);

						/* for nets with no physical representation ... */
						edifgbl->current_arc = NOARCINST;
						if (lx == hx && ly == hy)
						{
							if ((edifgbl->current_arc = newarcinst(ap, defaultarcwidth(ap),
								FIXANG|CANTSLIDE, fni, fpp, lx, ly, ni, pp, hx, hy, np)) == NOARCINST)
							{
								ttyputmsg("error, line #%d: could not create path (arc)",
									edifgbl->lineno);
							}
						}

						/* use unrouted connection for NETLIST views */
						else if (edifgbl->active_view == VNETLIST)
						{
							if ((edifgbl->current_arc = newarcinst(ap, defaultarcwidth(ap),
								CANTSLIDE, fni, fpp, lx, ly, ni, pp, hx, hy, np)) == NOARCINST)
							{
								ttyputmsg("error, line #%d: could not create auto-path",
									edifgbl->lineno);
							}
						}
						/* add the net name */
						if (edifgbl->current_arc != NOARCINST)
							io_ednamearc(edifgbl, edifgbl->current_arc);
					}
				}
				break;
			case KINSTANCEREF:
				break;
			case KPATH:
				/* check for openShape type path */
				if (edifgbl->path_width == 0 &&
					edifgbl->geometry != GNET && edifgbl->geometry != GBUS) goto dopoly;
				fcnt = 0;
				if (edifgbl->geometry == GBUS || edifgbl->isarray) np = sch_buspinprim;
				else np = sch_wirepinprim;
				if (edifgbl->points == NOEPT) break;
				for (point = edifgbl->points; point->nextpt != NOEPT; point = point->nextpt)
				{
					if (edifgbl->geometry == GNET || edifgbl->geometry == GBUS)
					{
						/* create a pin to pin connection */
						if (fcnt == 0)
						{
							/* look for the first pin */
							if ((fcnt = io_edfindport(edifgbl->current_cell, point->x,
								point->y, sch_wirearc, fnis, fpps)) == 0)
							{
								/* create the "from" pin */
								if ((fnis[0] = newnodeinst(np, point->x + np->lowx, point->x + np->highx,
									point->y + np->lowy, point->y + np->highy,
									0, 0, edifgbl->current_cell)) == NONODEINST)
										fcnt = 0;
								else
								{
									fpps[0] = (edifgbl->geometry == GBUS || edifgbl->isarray) ?
										default_busport : default_port;
									fcnt = 1;
								}
							}
						}
						/* now the second ... */
						if ((tcnt = io_edfindport(edifgbl->current_cell, point->nextpt->x,
							point->nextpt->y, sch_wirearc, tnis, tpps)) == 0)
						{
							/* create the "to" pin */
							if ((tnis[0] = newnodeinst(np, point->nextpt->x + np->lowx,
								point->nextpt->x + np->highx,
								point->nextpt->y + np->lowy,
								point->nextpt->y + np->highy,
								0, 0, edifgbl->current_cell)) == NONODEINST)
								tcnt = 0;
							else
							{
								tpps[0] = (edifgbl->geometry == GBUS || edifgbl->isarray) ?
									default_busport : default_port;
								tcnt = 1;
							}
						}

						if (tcnt == 0 || fcnt == 0)
						{
							ttyputerr("error, line #%d: could not create path", edifgbl->lineno);
							edifgbl->errors++;
						} else
						{
							/* connect it */
							for (count = 0; count < fcnt || count < tcnt; count++)
							{
								if (count < fcnt)
								{
									lastpin = fnis[count];
									lastport = fpps[count];

									/* check node for array variable */
									if ((var = getvalkey((INTBIG)lastpin, VNODEINST, VSTRING, EDIF_array_key)) != NOVARIABLE)
										fbus = 1;
									else if (lastport->protoname[strlen(lastport->protoname)-1] == ']')
										fbus = 1;
									else fbus = 0;
								}
								if (count < tcnt)
								{
									ni = tnis[count];
									pp = tpps[count];

									/* check node for array variable */
									if ((var = getvalkey((INTBIG)ni, VNODEINST, VSTRING, EDIF_array_key)) != NOVARIABLE)
										tbus = 1;
									else if (pp->protoname[strlen(pp->protoname)-1] == ']')
										tbus = 1;
									else tbus = 0;
								}

								/* if bus to bus */
								if ((lastport == default_busport || fbus) &&
									(pp == default_busport || tbus)) ap = sch_busarc;
										/* wire to other */
								else ap = sch_wirearc;

								if ((ai = newarcinst(ap, defaultarcwidth(ap), FIXANG|CANTSLIDE,
									lastpin, lastport, point->x, point->y,
									ni, pp, point->nextpt->x, point->nextpt->y,
									edifgbl->current_cell)) == NOARCINST)
								{
									ttyputerr("error, line #%d: could not create path (arc)",
										edifgbl->lineno);
									edifgbl->errors++;
								}
								else if (edifgbl->geometry == GNET && edifgbl->points == point)
								{
									if (edifgbl->net_reference)
									{
										var = setvalkey((INTBIG)ai, VARCINST, EDIF_name_key,
											(INTBIG)edifgbl->net_reference, VSTRING);
									}
									if (edifgbl->net_name)
									{
										/* set name of arc but don't display name */
										var = setvalkey((INTBIG)ai, VARCINST, el_arc_name,
											(INTBIG)edifgbl->net_name, VSTRING);
									}
								}
								else if (edifgbl->geometry == GBUS && edifgbl->points == point)
								{
									if (edifgbl->bundle_reference)
									{
										var = setvalkey((INTBIG)ai, VARCINST, EDIF_name_key,
											(INTBIG)edifgbl->bundle_reference, VSTRING);
									}
									if (edifgbl->bundle_name)
									{
										/* set bus' EDIF name but don't display name */
										var = setvalkey((INTBIG)ai, VARCINST, el_arc_name,
											(INTBIG)edifgbl->bundle_name, VSTRING);
									}
								}
							}
							if (ai != NOARCINST) edifgbl->current_arc = ai;
							for (count = 0; count < tcnt; count++)
							{
								fnis[count] = tnis[count];
								fpps[count] = tpps[count];
							}
							fcnt = tcnt;
						}
					} else
					{
						/* rectalinear paths with some width */
						/* create a path from here to there (orthogonal only now) */
						lx = hx = point->x;
						ly = hy = point->y;
						if (lx > point->nextpt->x) lx = point->nextpt->x;
						if (hx < point->nextpt->x) hx = point->nextpt->x;
						if (ly > point->nextpt->y) ly = point->nextpt->y;
						if (hy < point->nextpt->y) hy = point->nextpt->y;
						if (ly == hy || edifgbl->extend_end)
						{
							ly -= edifgbl->path_width/2;
							hy += edifgbl->path_width/2;
						}
						if (lx == hx || edifgbl->extend_end)
						{
							lx -= edifgbl->path_width/2;
							hx += edifgbl->path_width/2;
						}
						ni = newnodeinst(edifgbl->figure_group, lx, hx, ly, hy,
							io_edgettrans(edifgbl->orientation),
							io_edgetrot(edifgbl->orientation),
							edifgbl->current_cell);
						if (ni == NONODEINST)
						{
							ttyputerr("error, line #%d: could not create path",edifgbl->lineno);
							edifgbl->errors++;
						}
					}
				}
				io_edfreeptlist(edifgbl);
				break;
			case KCIRCLE:
				if (edifgbl->pt_count == 2 )
				{
					lx = mini(edifgbl->points->x, edifgbl->lastpt->x);
					hx = maxi(edifgbl->points->x, edifgbl->lastpt->x);
					ly = mini(edifgbl->points->y, edifgbl->lastpt->y);
					hy = maxi(edifgbl->points->y, edifgbl->lastpt->y);
					if (lx == hx)
					{
						lx -= (hy - ly)>>1;
						hx += (hy - ly)>>1;
					} else
					{
						ly -= (hx - lx)>>1;
						hy += (hx - lx)>>1;
					}

					/* create the node instance */
					ni = newnodeinst(art_circleprim, lx, hx, ly, hy,
						io_edgettrans(edifgbl->orientation),
						io_edgetrot(edifgbl->orientation),
						edifgbl->current_cell);
					if (ni == NONODEINST)
					{
						ttyputerr("error, line #%d: could not create circle",
							edifgbl->lineno);
						edifgbl->errors++;
					}
				}
				io_edfreeptlist(edifgbl);
				break;
			case KNAME:
				/* save the data and break */
				(void)strcpy(edifgbl->save_text.string, edifgbl->string);
				edifgbl->string[0] = '\0';
				edifgbl->save_text.points = edifgbl->points;
				edifgbl->points = NOEPT;
				edifgbl->save_text.pt_count = edifgbl->pt_count;
				edifgbl->pt_count = 0;
				edifgbl->save_text.textheight = edifgbl->textheight;
				edifgbl->textheight = 0;
				edifgbl->save_text.justification = edifgbl->justification;
				edifgbl->justification = LOWERLEFT;
				edifgbl->save_text.orientation = edifgbl->orientation;
				edifgbl->orientation = OR0;
				edifgbl->save_text.visible = edifgbl->visible;
				edifgbl->visible = 1;
				break;
			case KSTRINGDISPLAY:
				if (kstack_ptr <= 1)
				{
					ttyputerr("error, line #%d: bad location for \"stringDisplay\"",
						edifgbl->lineno);
					edifgbl->errors++;
				}
				else if (kstack[kstack_ptr-1] == KRENAME)
				{
					/* save the data and break */
					(void)strcpy(edifgbl->save_text.string, edifgbl->string);
					edifgbl->string[0] = '\0';
					edifgbl->save_text.points = edifgbl->points;
					edifgbl->points = NOEPT;
					edifgbl->save_text.pt_count = edifgbl->pt_count;
					edifgbl->pt_count = 0;
					edifgbl->save_text.textheight = edifgbl->textheight;
					edifgbl->textheight = 0;
					edifgbl->save_text.justification = edifgbl->justification;
					edifgbl->justification = LOWERLEFT;
					edifgbl->save_text.orientation = edifgbl->orientation;
					edifgbl->orientation = OR0;
					edifgbl->save_text.visible = edifgbl->visible;
					edifgbl->visible = 1;
				}

				/* output the data for annotate (display graphics) and string (on properties) */
				else if (kstack[kstack_ptr-1] == KANNOTATE ||
					kstack[kstack_ptr-1] == KSTRING)
				{
					/* see if a pre-existing node or arc exists to add text */
					ai = NOARCINST;
					ni = NONODEINST;
					if (edifgbl->property_reference[0] != '\0' && edifgbl->current_node != NONODEINST)
					{
						ni = edifgbl->current_node;
						key = makekey(edifgbl->property_reference);
					}
					else if (edifgbl->property_reference[0] != '\0' && edifgbl->current_arc != NOARCINST)
					{
						ai = edifgbl->current_arc;
						key = makekey(edifgbl->property_reference);
					} else
					{
						/* create the node instance */
						ni = newnodeinst(gen_invispinprim,
							edifgbl->points->x, edifgbl->points->x,
							edifgbl->points->y, edifgbl->points->y, 0, 0,
							edifgbl->current_cell);
						key = EDIF_annotate_key;
					}
					if (ni != NONODEINST || ai != NOARCINST)
					{
						if (ni != NONODEINST)
						{
							var = setvalkey((INTBIG)ni, VNODEINST, key, (INTBIG)edifgbl->string,
								(edifgbl->visible ? VSTRING|VDISPLAY : VSTRING));

							/* now set the position, relative to the center of the current object */
							xoff = edifgbl->points->x - ((ni->highx+ni->lowx)>>1);
							yoff = edifgbl->points->y - ((ni->highy+ni->lowy)>>1);
						} else
						{
							var = setvalkey((INTBIG)ai, VARCINST, key, (INTBIG)edifgbl->string,
								(edifgbl->visible ? VSTRING|VDISPLAY : VSTRING));

							/* now set the position, relative to the center of the current object */
							xoff = edifgbl->points->x - ((ai->end[0].xpos + ai->end[1].xpos)>>1);
							yoff = edifgbl->points->y - ((ai->end[0].ypos + ai->end[1].ypos)>>1);
						}

						/* convert to quarter lambda units */
						xoff = 4*xoff / edifgbl->technology->deflambda;
						yoff = 4*yoff / edifgbl->technology->deflambda;

						/* determine the size of text, 0.0278 in == 2 points or 36 (2xpixels) == 1 in
		 				  fonts range from TXT4P to TXT20P, in 2 pixel increments */
						if (edifgbl->textheight == 0) i = TXTSMALL; else
						{
							i = (int)((double)edifgbl->textheight * 3.6 /
								(double)edifgbl->technology->deflambda + 0.5) - 2;
							if (i < TXT4P) i = TXT4P;
							if (i > TXT20P) i = TXT20P;
						}

						descript = var->textdescript;
						descript &= ~(VTPOSITION|VTXOFF|VTYOFF|VTSIZE);
						descript |= (abs(xoff) << VTXOFFSH) | (xoff < 0 ? VTXOFFNEG : 0) |
							(abs(yoff) << VTYOFFSH) | (yoff < 0 ? VTYOFFNEG : 0) |
							(i << VTSIZESH);
						switch (edifgbl->justification)
						{
							case UPPERLEFT:
								descript |=  VTPOSUPRIGHT;
								break;
							case UPPERCENTER:
								descript |=  VTPOSUP;
								break;
							case UPPERRIGHT:
								descript |=  VTPOSUPLEFT;
								break;
							case CENTERLEFT:
								descript |=  VTPOSRIGHT;
								break;
							case CENTERCENTER:
								descript |=  VTPOSCENT;
								break;
							case CENTERRIGHT:
								descript |=  VTPOSLEFT;
								break;
							case LOWERLEFT:
								descript |=  VTPOSDOWNRIGHT;
								break;
							case LOWERCENTER:
								descript |=  VTPOSDOWN;
								break;
							case LOWERRIGHT:
								descript |=  VTPOSDOWNLEFT;
								break;
						}
						var->textdescript = descript;
					} else
					{
						ttyputerr("error, line #%d: nothing to attach text to", edifgbl->lineno);
						edifgbl->errors++;
					}
				}

				/* clean up DISPLAY attributes */
				io_edfreeptlist(edifgbl);
				edifgbl->cur_nametbl = NONAMETABLE;
				edifgbl->visible = 1;
				edifgbl->justification = LOWERLEFT;
				edifgbl->textheight = 0;
				break;
			case KDOT:
				if (edifgbl->geometry == GPIN)
				{
					for (eport = edifgbl->ports; eport != NOEDPORT; eport = eport->next)
					{
						if (!namesame(eport->reference, edifgbl->name)) break;
					}
					if (eport != NOEDPORT)
					{
						/* create a internal wire port using the pin-proto, and create the port
	  					 and export */
						ni = newnodeinst(edifgbl->proto,
							edifgbl->points->x+edifgbl->proto->lowx,
							edifgbl->points->x+edifgbl->proto->highx,
							edifgbl->points->y+edifgbl->proto->lowy,
							edifgbl->points->y+edifgbl->proto->highy,
							io_edgettrans(edifgbl->orientation),
							io_edgetrot(edifgbl->orientation),
							edifgbl->current_cell);
						if (ni == NONODEINST)
						{
							ttyputerr("error, line #%d: could not create pin", edifgbl->lineno);
							edifgbl->errors++;
						}
						(void)strcpy(portname, eport->name);
						for (dup = 0, (void)strcpy(basename, portname);
							(ppt = getportproto(ni->parent, portname)) != NOPORTPROTO; dup++)
						{
							if (dup == 0) pp = ppt;
							(void)sprintf(portname, "%s_%d", basename, dup+1);
						}
						if ((ppt = newportproto(ni->parent, ni, default_iconport, portname)) == NOPORTPROTO)
						{
							ttyputerr("error, line #%d: could not create port <%s>",
								edifgbl->lineno, portname);
							edifgbl->errors++;
						} else
						{
							/* set the direction */
							switch (eport->direction)
							{
								case INPUT:
									ppt->userbits = (ppt->userbits & ~STATEBITS) | INPORT;
									break;
								case OUTPUT:
									ppt->userbits = (ppt->userbits & ~STATEBITS) | OUTPORT;
									break;
								case INOUT:
									ppt->userbits = (ppt->userbits & ~STATEBITS) | BIDIRPORT;
									break;
							}
						}
					}
				} else
				{
					/* create the node instance */
					ni = newnodeinst(edifgbl->figure_group != NONODEPROTO ?
						edifgbl->figure_group : art_boxprim,
						edifgbl->points->x, edifgbl->points->x,
						edifgbl->points->y, edifgbl->points->y,
						io_edgettrans(edifgbl->orientation),
						io_edgetrot(edifgbl->orientation),
						edifgbl->current_cell);
					if (ni == NONODEINST)
					{
						ttyputerr("error, line #%d: could not create rectangle",
							edifgbl->lineno);
						edifgbl->errors++;
					}
				}
				io_edfreeptlist(edifgbl);
				break;
			case KRECTANGLE:
				if (kstack_ptr > 1 && (kstack[kstack_ptr-1] == KPAGESIZE ||
					kstack[kstack_ptr-1] == KBOUNDINGBOX)) break;
				if (edifgbl->pt_count == 2)
				{
					/* create the node instance */
					if (edifgbl->points->x > edifgbl->lastpt->x)
					{
						lx = edifgbl->lastpt->x;
						hx = edifgbl->points->x;
					} else
					{
						hx = edifgbl->lastpt->x;
						lx = edifgbl->points->x;
					}
					if (edifgbl->points->y > edifgbl->lastpt->y)
					{
						ly = edifgbl->lastpt->y;
						hy = edifgbl->points->y;
					} else
					{
						hy = edifgbl->lastpt->y;
						ly = edifgbl->points->y;
					}
					ni = newnodeinst(edifgbl->figure_group != NONODEPROTO ?
						edifgbl->figure_group : art_boxprim,
						lx, hx, ly, hy,
						io_edgettrans(edifgbl->orientation),
						io_edgetrot(edifgbl->orientation),
						edifgbl->current_cell);
					if (ni == NONODEINST)
					{
						ttyputerr("error, line #%d: could not create rectangle",
							edifgbl->lineno);
						edifgbl->errors++;
					}
					else if (edifgbl->figure_group == art_openeddottedpolygonprim)
					{
						cnt = 5;
						cx = (edifgbl->points->x + edifgbl->lastpt->x) / 2;
						cy = (edifgbl->points->y + edifgbl->lastpt->y) / 2;
						pts[0] = edifgbl->points->x-cx;
						pts[1] = edifgbl->points->y-cy;
						pts[2] = edifgbl->points->x-cx;
						pts[3] = edifgbl->lastpt->y-cy;
						pts[4] = edifgbl->lastpt->x-cx;
						pts[5] = edifgbl->lastpt->y-cy;
						pts[6] = edifgbl->lastpt->x-cx;
						pts[7] = edifgbl->points->y-cy;
						pts[8] = edifgbl->points->x-cx;
						pts[9] = edifgbl->points->y-cy;

						/* store the trace information */
						(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)pts,
							VINTEGER|VISARRAY|((cnt*2)<<VLENGTHSH));
					}
					else if (edifgbl->geometry == GPIN)
					{
						/* create a rectangle using the pin-proto, and create the port
	   					 * and export
	   					 */
						/* ensure full sized port */
						lx += edifgbl->proto->lowx;
						hx += edifgbl->proto->highx;
						ly += edifgbl->proto->lowy;
						hy += edifgbl->proto->highy;

						/* create the node instance */
						ni = newnodeinst(edifgbl->proto, lx, hx, ly, hy,
							io_edgettrans(edifgbl->orientation),
							io_edgetrot(edifgbl->orientation),
							edifgbl->current_cell);
						if (ni == NONODEINST)
						{
							ttyputerr("error, line #%d: could not create pin",
								edifgbl->lineno);
							edifgbl->errors++;
						}
						if ((ppt = getportproto(ni->parent, edifgbl->name)) == NOPORTPROTO &&
							(ppt = newportproto(ni->parent, ni, default_iconport, edifgbl->name))
							== NOPORTPROTO)
						{
							ttyputerr("error, line #%d: could not create port <%s>",
								edifgbl->lineno, edifgbl->name);
							edifgbl->errors++;
						}
					}
				}
				io_edfreeptlist(edifgbl);

				break;

			case KARC:
				/* set array values */
				for (point = edifgbl->points, i = 0; point && i < 3; point = point->nextpt,i++)
				{
					x[i] = point->x;
					y[i] = point->y;
				}
				io_ededif_arc(x, y, &cx, &cy, &radius, &so, &ar, &j, &trans);
				lx = cx - radius;   hx = cx + radius;
				ly = cy - radius;   hy = cy + radius;

				/* get the bounds of the circle */
				ni = newnodeinst(edifgbl->figure_group != NONODEPROTO &&
					edifgbl->figure_group != art_boxprim ?
					edifgbl->figure_group : art_circleprim, lx, hx, ly, hy,
					(INTSML)trans, (INTSML)j, edifgbl->current_cell);
				if (ni == NONODEINST)
				{
					ttyputerr("error, line #%d: could not create arc", edifgbl->lineno);
					edifgbl->errors++;
				} else
				{
					/* store the angle of the arc */
					setarcdegrees(ni, so, ar*EPI/1800.0);
				}
				io_edfreeptlist(edifgbl);
				break;

			case KSYMBOL:
			case KPAGE:
				if ((np = edifgbl->current_cell) != NONODEPROTO)
				{
					/* now compute the bounds of this cell */
					db_boundfacet(np, &np->lowx, &np->highx, &np->lowy, &np->highy);
				}
				edifgbl->active_view = VNULL;
				break;

			case KOPENSHAPE:
			case KPOLYGON:
dopoly:
				if (edifgbl->pt_count == 0) break;

				/* get the bounds of the poly */
				lx = hx = edifgbl->points->x;
				ly = hy = edifgbl->points->y;
				point = edifgbl->points->nextpt;
				while (point)
				{
					if (lx > point->x) lx = point->x;
					if (hx < point->x) hx = point->x;
					if (ly > point->y) ly = point->y;
					if (hy < point->y) hy = point->y;
					point = point->nextpt;
				}
				if (lx != hx || ly != hy)
				{
					if (edifgbl->figure_group != NONODEPROTO && edifgbl->figure_group != art_boxprim)
						np = edifgbl->figure_group; else
					{
						if (edifgbl->state == KPOLYGON) np = art_closedpolygonprim; else
							np = art_openedpolygonprim;
					}
					ni = newnodeinst(np, lx, hx, ly, hy, io_edgettrans(edifgbl->orientation),
						io_edgetrot(edifgbl->orientation), edifgbl->current_cell);
					if (ni == NONODEINST)
					{
						ttyputerr("error, line #%d: could not create polygon", edifgbl->lineno);
						edifgbl->errors++;
					} else
					{
						pt = trace = emalloc((edifgbl->pt_count*2*SIZEOFINTBIG), el_tempcluster);
						if (trace == 0) RET_NOMEMORY();
						cx = (hx + lx) / 2;
						cy = (hy + ly) / 2;
						point = edifgbl->points;
						while (point != NOEPT)
						{
							*pt++ = point->x - cx;
							*pt++ = point->y - cy;
							point = point->nextpt;
						}

						/* store the trace information */
						(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)trace,
							VINTEGER|VISARRAY|((edifgbl->pt_count*2)<<VLENGTHSH));

						/* free the polygon memory */
						efree((char *)trace);
					}
				}
				io_edfreeptlist(edifgbl);
				break;
			case KARRAY:
				switch (kstack[kstack_ptr-1])
				{
					case KPORT:
						io_edmake_electric_name(edifgbl->port_reference, edifgbl->port_name, edifgbl->arrayx,
							edifgbl->arrayy);
				}
				if (edifgbl->arrayx == 0) edifgbl->arrayx = 1;
				if (edifgbl->arrayy == 0) edifgbl->arrayy = 1;
				break;
			case KMEMBER:
				if (edifgbl->memberx != -1)
				{
					/* adjust the name of the current INSTANCE/NET/PORT(/VALUE) */
					if (edifgbl->membery != -1)
						(void)sprintf(basename, "[%d,%d]", edifgbl->memberx, edifgbl->membery);
					else (void)sprintf(basename, "[%d]", edifgbl->memberx);
					switch (kstack[kstack_ptr-1])
					{
						case KINSTANCEREF:
							(void)strcat(edifgbl->instance_reference, basename);
							break;
						case KPORTREF:
							(void)strcat(edifgbl->port_reference, basename);
							break;
					}
				}
				break;
			case KDESIGN:
				if (edifgbl->proto != NONODEPROTO)
				{
					if (io_verbose < 0)
					{
						(void)initinfstr();
						(void)addstringtoinfstr("Design ");
						(void)addstringtoinfstr(edifgbl->proto->cell->cellname);
						DiaSetText(2, returninfstr());
					}
					else ttyputmsg("Design %s", edifgbl->proto->cell->cellname);
					edifgbl->library->curnodeproto = edifgbl->proto;
				}
				break;

			case KEDIF:
				/* free the port freelist */
				for (eport = edifgbl->free_ports; eport != NOEDPORT; eport = neport)
				{
					neport = eport->next;
					efree((char *)eport);
				}
				edifgbl->free_ports = NOEDPORT;

				/* free the netport freelist */
				for (nport = edifgbl->free_netports; nport != NOEDNETPORT; nport = nnport)
				{
					nnport = nport->next;
					efree((char *)nport);
				}
				edifgbl->free_netports = NOEDNETPORT;
				break;
			default:
				break;
			}
		}
		if (kstack_ptr) edifgbl->state = kstack[--kstack_ptr]; else
			edifgbl->state = KINIT;
		return 0;
	}
	else return 1;
}

/* module: io_edprocess_integer
   function: will do cleanup processing of an integer argument such as in (array (...) 1 2)
 */
void io_edprocess_integer(EDIFGBL_PTR edifgbl, INTBIG value)
{
	if (kstack_ptr)
	{
		if (!edifgbl->ignoreblock)
		{
			switch (edifgbl->state)
			{
				case KARRAY:
					if (edifgbl->arrayx == 0) edifgbl->arrayx = value;
					else if (edifgbl->arrayy == 0) edifgbl->arrayy = value;
					break;
				case KMEMBER:
					if (edifgbl->memberx == -1) edifgbl->memberx = value;
					else if (edifgbl->membery == -1) edifgbl->membery = value;
					break;
			}
		}
	}
}

INTSML io_edgetrot(OTYPES orientation)
{
	switch (orientation)
	{
		case OR0:     return 0;
		case OR90:    return 90*10;
		case OR180:   return 180*10;
		case OR270:   return 270*10;
		case OMX:     return 270*10;
		case OMY:     return 90*10;
		case OMXR90:  return 180*10;
		case OMYR90:  return 0;
	}
	return 0;
}

INTSML io_edgettrans(OTYPES orientation)
{
	switch (orientation)
	{
		case OR0:     return 0;
		case OR90:    return 0;
		case OR180:   return 0;
		case OR270:   return 0;
		case OMY:     return 1;
		case OMX:     return 1;
		case OMYR90:  return 1;
		case OMXR90:  return 1;
	}
	return 0;
}

INTBIG io_edclockwise(EPT_PTR pt1, EPT_PTR pt2, EPT_PTR pt3)
{
	INTBIG area;

	/* calculate value of x1*y2 + x2*y3 + x3*y1 - y1*x2 - y2*x3 - y3*x1 (area) */
	area = pt1->x * pt2->y + pt2->x * pt3->y + pt3->x * pt1->y -
		pt1->y * pt2->x - pt2->y * pt3->x - pt3->y * pt1->x;
	if (area > 0) return 0;
	return 1;
}

INTBIG io_edcompare_name(NAMETABLE_PTR *name1, NAMETABLE_PTR *name2)
{
	return namesame((*name1)->original, (*name2)->original);
}

/* Will check if a string is an valid integer */
INTBIG io_edis_integer(char *buffer)
{
	/* remove sign */
	if (*buffer == '+' || *buffer == '-') buffer++;
	while (isdigit(*buffer))
	{
		buffer++;
		if (*buffer == '\0') return 1;
	}
	return 0;
}

/* get a keyword routine */
char *io_edget_keyword(EDIFGBL_PTR edifgbl, char *buffer)
{
	char *p;

	/* look for a '(' before the edif keyword */
	while ((p = io_edget_token(edifgbl, buffer, 0)) && (*p != LEFTP))
	{
		if (*p == RIGHTP) io_edpop_stack(edifgbl); else
			if (io_edis_integer(buffer)) io_edprocess_integer(edifgbl, atoi(buffer));
	}
	return io_edget_token(edifgbl, buffer, 0);

}

/* get a token routine */
char *io_edget_token(EDIFGBL_PTR edifgbl, char *buffer, char idelim)
{
	char *sptr;
	char delim;

	/* locate the first non-white space character for non-delimited searches */
	if (!(delim = idelim))
	{
		if (io_edpos_token(edifgbl) == NULL)
			return NULL;
	}

	/* set up the string copy */
	sptr = buffer;
	if (!delim && *edifgbl->pos == '"')
	{
		/* string search */
		delim = '"';
		*(sptr++) = *(edifgbl->pos++);
	}

	/* non locate the next white space or the delimiter */
	while (1)
	{
		if (*edifgbl->pos == '\0')
		{
			/* end of string, get the next line */
			if (io_edget_line(edifgbl->buffer, LINE, edifgbl->edif_file) == 0)
			{
				/* end-of-file, if no delimiter return NULL, otherwise the string */
				if (delim) return NULL; else
				{
					*sptr = '\0';
					return buffer;
				}
			}
			edifgbl->pos = edifgbl->buffer;
		}
		else if (!delim)
		{
			switch (*edifgbl->pos)
			{
				/* the space characters */
				case NEWLINE:
				case RETURN:
					edifgbl->lineno++;
				case SPACE:
				case TAB:
					*sptr = '\0';
					/* skip to the next character */
					edifgbl->pos++;
					return buffer;
					/* special EDIF delimiters */
				case LEFTP:
				case RIGHTP:
					if (sptr == buffer)
						*(sptr++) = *(edifgbl->pos++);
					*sptr = '\0';
					return buffer;
				default:
					*(sptr++) = *(edifgbl->pos++);
					break;
			}
		} else
		{
			/* check for user specified delimiter */
			if (*edifgbl->pos == delim)
			{
				/* skip the delimiter, unless string */
				if (delim != idelim) *(sptr++) = *(edifgbl->pos++); else
					edifgbl->pos++;
				*sptr = '\0';
				return buffer;
			} else
			{
				if (*edifgbl->pos == RETURN || *edifgbl->pos == NEWLINE)
				{
					edifgbl->lineno++;
				}

				*(sptr++) = *(edifgbl->pos++);
			}
		}
	}
}

/*
 * Module: io_edpos_token
 * Function: special routine to position to the next token
 */
char *io_edpos_token(EDIFGBL_PTR edifgbl)
{
	while (1)
	{
		switch (*edifgbl->pos)
		{
			case '\0':
				/* end of string, get the next line */
				if (io_edget_line(edifgbl->buffer, LINE, edifgbl->edif_file) == 0)
				{
					/* end-of-file */
					return NULL;
				}
				edifgbl->pos = edifgbl->buffer;
				break;

				/* the space characters */
			case NEWLINE:
			case RETURN:
				edifgbl->lineno++;
			case SPACE:
			case TAB:
				edifgbl->pos++;
				break;
			default:
				return edifgbl->pos;
		}
	}
}

/*
 * Module: io_edflush_line
 * Function: special routine to position to the next delimeter or
 * line
 */
char *io_edflush_line(EDIFGBL_PTR edifgbl, char delim)
{
	/* locate to the next delimeter or end-of-line */
	while (*edifgbl->pos != delim)
	{
		if (*edifgbl->pos == '\0')
		{
			/* get the next line */
			if (io_edget_line(edifgbl->buffer, LINE, edifgbl->edif_file) == 0)
			{
				/* end-of-file */
				return NULL;
			}
			edifgbl->pos = edifgbl->buffer;
		}
		else if (*edifgbl->pos == RETURN || *edifgbl->pos == NEWLINE)
		{
			edifgbl->lineno++;
			edifgbl->pos++;
		}
		else
			edifgbl->pos++;
	}
	return edifgbl->pos;
}

/*
 * routine to read a line from file "file" and place it in "line" (which has
 * up to "limit" characters).  Returns 0 on end-of-file.  Updates "filepos".
 */
INTSML io_edget_line(char *line, INTSML limit, FILE *file)
{
	REGISTER char *pp;
	REGISTER INTSML c, total;

	pp = line;
	total = 1;
	for(;;)
	{
		c = xgetc(file);
		if (c == EOF)
		{
			if (pp == line) return(0);
			break;
		}
		filepos++;
		if ((*pp = (char)c) == '\n')
		{
			pp++;
			break;
		}
		pp++;
		if ((++total) >= limit) break;
	}
	*pp = 0;
	return(1);
}

/* get a delimeter routine */
void io_edget_delim(EDIFGBL_PTR edifgbl, char delim)
{
	char *token;

	if ((token = io_edpos_token(edifgbl)) == NULL)
		longjmp(edifgbl->env, (int)"unexpected end-of-file");
	if (*token != delim)
		longjmp(edifgbl->env, (int)"illegal delimeter");
	else edifgbl->pos++;

}

/************ INPUT SUPPORT *********/

NODEINST *io_edreplacenodeinst(NODEINST *ni, NODEPROTO *np, PORTPROTO *pp, INTBIG x, INTBIG y)
{
	PORTARCINST *pi;
	ARCINST *ai, *newai;
	NODEINST *endnodeinst[2], *newno;
	PORTPROTO *endportproto[2];
	INTBIG endx[2], endy[2], i;

	/* first create the new nodeinst in place */
	newno = newnodeinst(np, x + np->lowx, x + np->highx,
		y + np->lowy, y + np->highy, ni->transpose,
		ni->rotation, ni->parent);
	if (newno == NONODEINST) return(NONODEINST);

	/* now replace the arcs */
	for (pi = ni->firstportarcinst; pi != NOPORTARCINST; pi = pi->nextportarcinst)
	{
		ai = pi->conarcinst;
		for (i=0; i<2; i++)
		{
			if (ai->end[i].nodeinst == ni)
			{
				endnodeinst[i] = newno;
				endportproto[i] = pp;
				endx[i] = ai->end[i].xpos;
				endy[i] = ai->end[i].ypos;
			} else
			{
				endnodeinst[i] = ai->end[i].nodeinst;
				endportproto[i] = ai->end[i].portarcinst->proto;
				endx[i] = ai->end[i].xpos;
				endy[i] = ai->end[i].ypos;
			}
		}

		/* replace the arc with another arc */
		newai = newarcinst(ai->proto, ai->width, ai->userbits,
			endnodeinst[0], endportproto[0], endx[0],endy[0],
			endnodeinst[1], endportproto[1], endx[1],endy[1],
			ni->parent);
		if (newai == NOARCINST) return(NONODEINST);
		if (copyvars((INTBIG)ai, VARCINST, (INTBIG)newai, VARCINST) != 0)
			return NONODEINST;
		for (i=0; i<2; i++)
			if (copyvars((INTBIG)ai->end[i].portarcinst, VPORTARCINST,
				(INTBIG)newai->end[i].portarcinst, VPORTARCINST) != 0)
				return NONODEINST;
		killarcinst(ai);
	}
	killnodeinst(ni);
	return newno;
}

char *io_edstralloc(char *str)
{
	char *newmem;
	unsigned int len;

	len = strlen(str);
	if ((newmem = (char *)emalloc(len + 1, io_aid->cluster)) == NULL) return NULL;
	(void)strcpy(newmem, str);
	return newmem;
}

/* This function allocates ports */
INTBIG io_edallocport(EDIFGBL_PTR edifgbl)
{
	EDPORT_PTR port;

	if (edifgbl->free_ports != NOEDPORT)
	{
		port = edifgbl->free_ports;
		edifgbl->free_ports = port->next;
	}
	else if ((port = (EDPORT_PTR) emalloc((unsigned)sizeof (EDPORT), io_aid->cluster)) == NOEDPORT)
		RET_NOMEMORY();

	port->next = edifgbl->ports;
	edifgbl->ports = port;
	if ((port->reference = io_edstralloc(edifgbl->port_reference)) == NULL) RET_NOMEMORY();
	if ((port->name = io_edstralloc(edifgbl->port_name)) == NULL) RET_NOMEMORY();

	port->direction = edifgbl->direction;
	port->arrayx = edifgbl->arrayx;
	port->arrayy = edifgbl->arrayy;
	return 0;
}

INTBIG io_edallocproperty(EDIFGBL_PTR edifgbl, char *name, PROPERTY_TYPE type, int integer,
	float number, char *string)
{
	EDPROPERTY_PTR property;

	if (edifgbl->free_properties != NOEDPROPERTY)
	{
		property = edifgbl->free_properties;
		edifgbl->free_properties = property->next;
	} else
		if ((property = (EDPROPERTY_PTR) emalloc((unsigned)sizeof (EDPROPERTY), io_aid->cluster)) == NOEDPROPERTY)
			RET_NOMEMORY();

	property->next = edifgbl->properties;
	edifgbl->properties = property;
	if ((property->name = io_edstralloc(name)) == NULL) RET_NOMEMORY();
	switch ((property->type = type))
	{
		case PINTEGER:
			property->val.integer = integer;
			break;
		case PNUMBER:
			property->val.number = number;
			break;
		case PSTRING:
			if ((property->val.string = io_edstralloc(string)) == NULL) RET_NOMEMORY();
			break;
	}
	return 0;

}

/* This function allocates net ports */
INTBIG io_edallocnetport(EDIFGBL_PTR edifgbl)
{
	EDNETPORT_PTR netport;

	if (edifgbl->free_netports != NOEDNETPORT)
	{
		netport = edifgbl->free_netports;
		edifgbl->free_netports = netport->next;
	} else
		if ((netport = (EDNETPORT_PTR)emalloc((unsigned)sizeof (EDNETPORT), io_aid->cluster)) == NOEDNETPORT)
			RET_NOMEMORY();

	if (edifgbl->firstnetport == NOEDNETPORT) edifgbl->firstnetport = netport; else
		edifgbl->lastnetport->next = netport;

	edifgbl->lastnetport = netport;
	netport->next = NOEDNETPORT;
	netport->ni = NONODEINST;
	netport->pp = NOPORTPROTO;
	netport->member = 0;
	return 0;
}

/*
 * Module: io_edeq_of_a_line
 * Function: Calculates the equation of a line (Ax + By + C = 0)
 * Inputs:  sx, sy  - Start point
 *          ex, ey  - End point
 *          px, py  - Point line should pass through
 * Outputs: A, B, C - constants for the line
 */
void io_edeq_of_a_line(double sx, double sy, double ex, double ey, double *A, double *B, double *C)
{
	if (sx == ex)
	{
		*A = 1.0;
		*B = 0.0;
		*C = -ex;
	} else if (sy == ey)
	{
		*A = 0.0;
		*B = 1.0;
		*C = -ey;
	} else
	{
		/* let B = 1 then */
		*B = 1.0;
		if (sx != 0.0)
		{
			/* Ax1 + y1 + C = 0    =>    A = -(C + y1) / x1 */
			/* C = -Ax2 - y2       =>    C = (Cx2 + y1x2) / x1 - y2   =>   Cx1 - Cx2 = y1x2 - y2x1  */
			/* C = (y2x1 - y1x2) / (x2 - x1) */
			*C = (ey * sx - sy * ex) / (ex - sx);
			*A = -(*C + sy) / sx;
		} else
		{
			*C = (sy * ex - ey * sx) / (sx - ex);
			*A = -(*C + ey) / ex;
		}
	}
}

/*
 * Module: io_eddetermine_intersection
 * Function: Will determine the intersection of two lines
 * Inputs: Ax + By + C = 0 form
 * Outputs: x, y - the point of intersection
 * returns 1 if found, 0 if non-intersecting
 */
INTBIG io_eddetermine_intersection(double A[2], double B[2], double C[2], double *x, double *y)
{
	double A1B2, A2B1;
	double A2C1, A1C2;
	double X, Y;

	/* check for parallel lines */
	if ((A1B2 = A[0] * B[1]) == (A2B1 = A[1] * B[0]))
	{
		/* check for coincident lines */
		if (C[0] == C[1]) return -1;
		return 0;
	}
	A1C2 = A[0] * C[1];
	A2C1 = A[1] * C[0];

	if (A[0])
	{
		Y = (A2C1 - A1C2) / (A1B2 - A2B1);
		X = -(B[0] * Y + C[0]) / A[0];
	} else
	{
		Y = (A1C2 - A2C1) / (A2B1 - A1B2);
		X = -(B[1] * Y + C[1]) / A[1];
	}
	*y = Y;
	*x = X;
	return 1;
}

void io_ededif_arc(INTBIG ix[3], INTBIG iy[3], INTBIG *ixc, INTBIG *iyc, INTBIG *r, double *so,
	double *ar, INTBIG *rot, INTBIG *trans)
{
	double x[3], y[3], px[3], py[3], a[3];
	double A[2], B[2], C[2];
	double dx, dy, R, area, xc, yc;
	INTSML i;

	for(i=0; i<3; i++)
	{
		x[i] = (double)ix[i];
		y[i] = (double)iy[i];
	}

	/* get line equations of perpendicular bi-sectors of p1 to p2 */
	px[1] = (x[0] + x[1]) / 2.0;
	py[1] = (y[0] + y[1]) / 2.0;

	/* now rotate end point 90 degrees */
	px[0] = px[1] - (y[0] - py[1]);
	py[0] = py[1] + (x[0] - px[1]);
	io_edeq_of_a_line(px[0], py[0], px[1], py[1], &A[0], &B[0], &C[0]);

	/* get line equations of perpendicular bi-sectors of p2 to p3 */
	px[1] = (x[2] + x[1]) / 2.0;
	py[1] = (y[2] + y[1]) / 2.0;

	/* now rotate end point 90 degrees */
	px[2] = px[1] - (y[2] - py[1]);
	py[2] = py[1] + (x[2] - px[1]);
	io_edeq_of_a_line(px[1], py[1], px[2], py[2], &A[1], &B[1], &C[1]);

	/* determine the point of intersection */
	(void)io_eddetermine_intersection(A, B, C, &xc, &yc);

	*ixc = rounddouble(xc);
	*iyc = rounddouble(yc);

	dx = ((double)*ixc) - x[0];
	dy = ((double)*iyc) - y[0];
	R = sqrt(dx * dx + dy * dy);
	*r = rounddouble(R);

	/* now calculate the angle to the start and endpoint */
	dx = x[0] - xc;  dy = y[0] - yc;
	if (dx == 0.0 && dy == 0.0)
	{
		ttyputerr("Domain error doing arc computation");
		return;
	}
	a[0] = atan2(dy, dx) * 1800.0 / EPI;
	if (a[0] < 0.0) a[0] += 3600.0;

	dx = x[2] - xc;  dy = y[2] - yc;
	if (dx == 0.0 && dy == 0.0)
	{
		ttyputerr("Domain error doing arc computation");
		return;
	}
	a[2] = atan2(dy, dx) * 1800.0 / EPI;
	if (a[2] < 0.0) a[2] += 3600.0;

	/* determine the angle of rotation and object orientation */
	/* determine the direction */
	/* calculate x1*y2 + x2*y3 + x3*y1 - y1*x2 - y2*x3 - y3*x1 */
	area = x[0]*y[1] + x[1]*y[2] + x[2]*y[0] - y[0]*x[1] - y[1]*x[2] - y[2]*x[0];
	if (area > 0.0)
	{
		/* counter clockwise */
		if (a[2] < a[0]) a[2] += 3600.0;
		*ar = a[2] - a[0];
		*rot = rounddouble(a[0]);
		*so = (a[0] - (double)*rot) * EPI / 1800.0;
		*trans = 0;
	} else
	{
		/* clockwise */
		if (a[0] > 2700)
		{
			*rot = 3600 - rounddouble(a[0]) + 2700;
			*so = ((3600.0 - a[0] + 2700.0) - (double)*rot) * EPI / 1800.0;
		} else
		{
			*rot = 2700 - rounddouble(a[0]);
			*so = ((2700.0 - a[0]) - (double)*rot) * EPI / 1800.0;
		}
		if (a[0] < a[2]) a[0] += 3600;
		*ar = a[0] - a[2];
		*trans = 1;
	}
}

/*
 * routine to generate an icon in library "lib" with name "iconname" from the
 * port list in "fpp".  The icon facet is called "pt" and it is constructed
 * from technology "tech".  The icon facet is returned (NONODEPROTO on error).
 * note: copied and adapted from us_makeiconfacet
 */
NODEPROTO *io_edmakeiconfacet(PORTPROTO *fpp, char *iconname, char *pt,
	TECHNOLOGY *tech, LIBRARY *lib)
{
	REGISTER NODEPROTO *np, *bbproto, *pinproto, *buspproto, *facetproto, *pintype;
	REGISTER NODEINST *bbni, *pinni;
	REGISTER PORTPROTO *pp, *port, *inputport, *outputport, *bidirport,
		*topport, *whichport, *bpp;
	REGISTER ARCPROTO *wireproto, *busproto, *wiretype;
	REGISTER INTSML inputside, outputside, bidirside, topside;
	REGISTER INTBIG character, index, xsize, ysize, xpos, ypos, xbbpos, ybbpos, spacing;
	REGISTER VARIABLE *var;

	/* get the necessary symbols */
	pinproto = getnodeproto("schematic:Icon_Pin");
	buspproto = getnodeproto("schematic:Icon_Pin");
	bbproto = getnodeproto("schematic:Bbox");
	facetproto = getnodeproto("generic:Facet-Center");
	wireproto = getarcproto("schematic:wire");
	busproto = getarcproto("schematic:bus");
	if (pinproto == NONODEPROTO || buspproto == NONODEPROTO ||
		bbproto == NONODEPROTO || facetproto == NONODEPROTO ||
		wireproto == NOARCPROTO || busproto == NOARCPROTO)
	{
		us_abortcommand("Cannot find necessary primitives");
		return(NONODEPROTO);
	}
	outputport = bbproto->firstportproto;
	topport = outputport->nextportproto;
	inputport = topport->nextportproto;
	bidirport = inputport->nextportproto;

	/* create the new icon facet */
	np = newnodeproto(pt, lib);
	if (np == NONODEPROTO)
	{
		us_abortcommand("Cannot create icon %s", pt);
		return(NONODEPROTO);
	}
	np->userbits |= WANTNEXPAND;

	/* determine number of inputs and outputs */
	inputside = outputside = bidirside = topside = 0;
	for(pp = fpp; pp != NOPORTPROTO; pp = pp->nextportproto)
	{
		if ((pp->userbits&BODYONLY) != 0) continue;
		character = pp->userbits & STATEBITS;

		/* special detection for power and ground ports */
		if (portispower(pp) != 0 || portisground(pp) != 0) character = GNDPORT;

		/* make a count of the types of ports and save it on the ports */
		switch (character)
		{
			case OUTPORT:
				pp->temp1 = outputside++;
				break;
			case BIDIRPORT:
				pp->temp1 = bidirside++;
				break;
			case PWRPORT:
			case GNDPORT:
				pp->temp1 = topside++;
				break;
			default:		/* INPORT, unlabeled, and all CLOCK ports */
				pp->temp1 = inputside++;
				break;
		}
	}

	/* create the Black Box with the correct size */
	ysize = maxi(maxi(inputside, outputside), 5) * 2 * tech->deflambda;
	xsize = maxi(maxi(topside, bidirside), 3) * 2 * tech->deflambda;

	/* create the Black Box instance */
	bbni = newnodeinst(bbproto, 0, xsize, 0, ysize, 0, 0, np);
	if (bbni == NONODEINST) return(NONODEPROTO);

	/* create the Facet Center instance */
	pinni = newnodeinst(facetproto, 0, 0, 0, 0, 0, 0, np);
	if (pinni == NONODEINST) return(NONODEPROTO);

	/* put the original cell name on the Black Box */
	var = setval((INTBIG)bbni, VNODEINST, "SCHEM_function", (INTBIG)iconname, VSTRING|VDISPLAY);
	if (var != NOVARIABLE) var->textdescript = defaulttextdescript(NOGEOM);

	/* place pins around the Black Box */
	for(pp = fpp; pp != NOPORTPROTO; pp = pp->nextportproto)
	{
		if ((pp->userbits&BODYONLY) != 0) continue;
		character = pp->userbits & STATEBITS;

		/* special detection for power and ground ports */
		if (portispower(pp) != 0 || portisground(pp) != 0) character = GNDPORT;

		/* make a count of the types of ports and save it on the ports */
		switch (character)
		{
			case OUTPORT:
				xpos = xsize + 2 * tech->deflambda;
				xbbpos = xsize;
				index = pp->temp1;
				spacing = 2 * tech->deflambda;
				if (outputside*2 < inputside) spacing = 4 * tech->deflambda;
				ybbpos = ypos = ysize - ((ysize - (outputside-1)*spacing) / 2 + index * spacing);
				whichport = outputport;
				break;
			case BIDIRPORT:
				index = pp->temp1;
				spacing = 2 * tech->deflambda;
				if (bidirside*2 < topside) spacing = 4 * tech->deflambda;
				xbbpos = xpos = xsize - ((xsize - (bidirside-1)*spacing) / 2 + index * spacing);
				ypos = -2 * tech->deflambda;
				ybbpos = 0;
				whichport = bidirport;
				break;
			case PWRPORT:
			case GNDPORT:
				index = pp->temp1;
				spacing = 2 * tech->deflambda;
				if (topside*2 < bidirside) spacing = 4 * tech->deflambda;
				xbbpos = xpos = xsize - ((xsize - (topside-1)*spacing) / 2 + index * spacing);
				ypos = ysize + 2 * tech->deflambda;
				ybbpos = ysize;
				whichport = topport;
				break;
			default:		/* INPORT, unlabeled, and all CLOCK ports */
				xpos = -2 * tech->deflambda;
				xbbpos = 0;
				index = pp->temp1;
				spacing = 2 * tech->deflambda;
				if (inputside*2 < outputside) spacing = 4 * tech->deflambda;
				ybbpos = ypos = ysize - ((ysize - (inputside-1)*spacing) / 2 + index * spacing);
				whichport = inputport;
				break;
		}

		/* determine type of pin */
		pintype = pinproto;
		wiretype = wireproto;
		if (pp->subnodeinst != NONODEINST)
		{
			bpp = pp;
			while (bpp->subnodeinst->proto->index == 0) bpp = bpp->subportproto;
			if (bpp->subnodeinst->proto == buspproto)
			{
				pintype = buspproto;
				wiretype = busproto;
			}
		}

		/* create the pin */
		pinni = newnodeinst(pintype, xpos, xpos, ypos, ypos, 0, 0, np);
		if (pinni == NONODEINST) return(NONODEPROTO);

		/* export the port that should be on this pin */
		port = newportproto(np, pinni, pintype->firstportproto, pp->protoname);
		if (port != NOPORTPROTO) port->userbits = pp->userbits | PORTDRAWN;

		/* wire this pin to the black box */
		(void)newarcinst(wiretype, defaultarcwidth(wiretype),
			us_makearcuserbits(wiretype), pinni, pintype->firstportproto,
			xpos, ypos, bbni, whichport, xbbpos, ybbpos, np);
	}
	return(np);
}

void io_ednamearc(EDIFGBL *edifgbl, ARCINST *ai)
{
	int Ix, Iy;
	char *name, basename[WORD+1] ;
	VARIABLE *var;

	if (edifgbl->isarray)
	{
		/* name the bus */
		if (edifgbl->net_reference)
		{
			var = setvalkey((INTBIG)ai, VARCINST, EDIF_name_key,
				(INTBIG)edifgbl->net_reference, VSTRING);
		}

		/* a typical foreign array bus will have some range value appended
		 * to the name. This will cause some odd names with duplicate values
		 */
		name = basename;
		*name = '\0';
		for (Ix = 0; Ix < edifgbl->arrayx; Ix++)
		{
			for (Iy = 0; Iy < edifgbl->arrayy; Iy++)
			{
				if (name != basename) *name++ = ',';
				if (edifgbl->arrayx > 1)
				{
					if (edifgbl->arrayy > 1)
						(void)sprintf(name, "%s[%d,%d]", edifgbl->net_name, Ix, Iy);
					else
						(void)sprintf(name, "%s[%d]", edifgbl->net_name, Ix);
				}
				else (void)sprintf(name, "%s[%d]", edifgbl->net_name, Iy);
				name += strlen(name);
			}
		}
		var = setvalkey((INTBIG)ai, VARCINST, el_arc_name, (INTBIG)basename, VSTRING);
	} else
	{
		if (edifgbl->net_reference)
		{
			var = setvalkey((INTBIG)ai, VARCINST, EDIF_name_key,
				(INTBIG)edifgbl->net_reference, VSTRING);
		}
		if (edifgbl->net_name)
		{
			/* set name of arc but don't display name */
			var = setvalkey((INTBIG)ai, VARCINST, el_arc_name,
				(INTBIG)edifgbl->net_name, VSTRING);
		}
	}
}

INTBIG io_edfindport(NODEPROTO *facet, INTBIG x, INTBIG y, ARCPROTO *ap, NODEINST *nis[],
	PORTPROTO *pps[])
{
	INTBIG cnt;
	ARCINST *ai = NOARCINST, *ar1,*ar2;
	ARCPROTO *nap;
	NODEINST *ni, *fno, *tno;
	PORTPROTO *fpt, *tpt, *pp;
	NODEPROTO *np, *pnt;
	INTBIG wid, bits1, bits2, lx, hx, ly, hy, j;
	INTBIG fendx, fendy, tendx, tendy;

	cnt = io_edfindport_geom(facet, x, y, ap, nis, pps, &ai);

	if (cnt == 0 && ai != NOARCINST)
	{
		/* direct hit on an arc, verify connection */
		nap = ai->proto;
		np = getpinproto(nap);
		if (np == NONODEPROTO) return 0;
		pp = np->firstportproto;
		for (j = 0; pp->connects[j] != NOARCPROTO; j++)
			if (pp->connects[j] == ap) break;
		if (pp->connects[j] == NOARCPROTO) return 0;

		/* try to split arc (from us_getnodeonarcinst)*/
		/* break is at (prefx, prefy): save information about the arcinst */
		fno = ai->end[0].nodeinst;	fpt = ai->end[0].portarcinst->proto;
		tno = ai->end[1].nodeinst;	tpt = ai->end[1].portarcinst->proto;
		fendx = ai->end[0].xpos;	  fendy = ai->end[0].ypos;
		tendx = ai->end[1].xpos;	  tendy = ai->end[1].ypos;
		wid = ai->width;  pnt = ai->parent;
		bits1 = bits2 = ai->userbits;
		if ((bits1&ISNEGATED) != 0)
		{
			if ((bits1&REVERSEEND) == 0) bits2 &= ~ISNEGATED; else
				bits1 &= ~ISNEGATED;
		}
		if (figureangle(fendx,fendy, x,y) != figureangle(x,y, tendx, tendy))
		{
			bits1 &= ~FIXANG;
			bits2 &= ~FIXANG;
		}

		/* create the splitting pin */
		lx = x - (np->highx-np->lowx)/2;   hx = lx + np->highx-np->lowx;
		ly = y - (np->highy-np->lowy)/2;   hy = ly + np->highy-np->lowy;
		ni = newnodeinst(np, lx,hx, ly,hy, 0, 0, pnt);
		if (ni == NONODEINST)
		{
			ttyputerr("Cannot create splitting pin");
			return 0;
		}
		endobjectchange((INTBIG)ni, VNODEINST);

		/* set the node, and port */
		nis[cnt] = ni;
		pps[cnt++] = pp;

		/* create the two new arcinsts */
		ar1 = newarcinst(nap, wid, bits1, fno, fpt, fendx, fendy, ni, pp, x, y, pnt);
		ar2 = newarcinst(nap, wid, bits2, ni, pp, x, y, tno, tpt, tendx, tendy, pnt);
		if (ar1 == NOARCINST || ar2 == NOARCINST)
		{
			ttyputerr("Error creating the split arc parts");
			return cnt;
		}
		(void)copyvars((INTBIG)ai, VARCINST, (INTBIG)ar1, VARCINST);
		endobjectchange((INTBIG)ar1, VARCINST);
		endobjectchange((INTBIG)ar2, VARCINST);

		/* delete the old arcinst */
		startobjectchange((INTBIG)ai, VARCINST);
		if (killarcinst(ai))
			ttyputerr("Error deleting original arc");
	}

	return cnt;
}

/* module: ro_mazefindport
   function: will locate the nodeinstance and portproto corresponding to
   to a direct intersection with the given point.
   inputs:
   facet - facet to search
   x, y  - the point to exam
   ap    - the arc used to connect port (must match pp)
   nis   - pointer to ni pointer buffer.
   pps   - pointer to portproto pointer buffer.
   outputs:
   returns cnt if found, 0 not found, -1 on error
   ni = found ni instance
   pp = found pp proto.
 */
INTBIG io_edfindport_geom(NODEPROTO *facet, INTBIG x, INTBIG y, ARCPROTO *ap,
	NODEINST *nis[], PORTPROTO *pps[], ARCINST **ai)
{
	REGISTER INTBIG j, cnt, sea;
	REGISTER GEOM *geom;
	static POLYGON *poly = NOPOLYGON;
	REGISTER PORTPROTO *pp;
	REGISTER NODEINST *ni;

	if (poly == NOPOLYGON) poly = allocpolygon(4, io_aid->cluster);

	cnt = 0;
	sea = initsearch(x, x, y, y, facet);
	for(;;)
	{
		geom = nextobject(sea);
		if (geom == NOGEOM) break;

		switch (geom->entrytype)
		{
			case OBJNODEINST:
				/* now locate a portproto */
				ni = geom->entryaddr.ni;
				for (pp = ni->proto->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
				{
					shapeportpoly(ni, pp, poly, 0);
					if (isinside(x, y, poly) != 0)
					{
						/* check if port connects to arc ...*/
						for (j = 0; pp->connects[j] != NOARCPROTO; j++)
							if (pp->connects[j] == ap)
						{
							nis[cnt] = ni;
							pps[cnt] = pp;
							cnt++;
						}
					}
				}
				break;

			case OBJARCINST :
				/* only accept valid wires */
				if (geom->entryaddr.ai->proto->tech == sch_tech)
					*ai = geom->entryaddr.ai;
				break;
		}
	}
	return cnt;
}

#endif  /* IOEDIF - at top */
